# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

GraphQL Middleware defined here also.

"""

from functools import partial
from inspect import isclass, iscoroutinefunction
import logging
from typing import TYPE_CHECKING, Any, Tuple, Union

from graphene.utils.str_converters import to_snake_case
from graphql.execution.utils import (
    get_operation_root_type, get_field_def
)
from graphql.execution.values import get_argument_values, get_variable_values
from graphql.language.base import parse, print_ast
from graphql.language import ast
from graphql.backend.base import GraphQLBackend, GraphQLDocument
from graphql.backend.core import execute_and_validate
from graphql.utils.base import type_from_ast
from graphql.type.definition import get_named_type
from promise import Promise
from rx import Observable

from cylc.flow.network.schema import NODE_MAP

if TYPE_CHECKING:
    from graphql.execution import ExecutionResult
    from graphql.language.ast import Document
    from graphql.type.schema import GraphQLSchema


logger = logging.getLogger(__name__)

STRIP_ARG = 'strip_null'
NULL_VALUE = None
EMPTY_VALUES: Tuple[list, dict] = ([], {})
STRIP_OPS = {'query', 'subscription'}


def grow_tree(tree, path, leaves=None):
    """Additively grows tree with leaves at terminal of new branch.

    Given existing dictionary, it follows the new path from root through
    existing trunk and without clobbering existing leaves when encountered.

    Args:
        tree (dict): Existing or new dictionary/tree.
        path (list): List of keys from root to branch end.
        leaves (dict, optional):
            Dictionary of information to put at path end.

    Returns:
        None
    """
    tree_loc = [tree, {}]
    for loc, key in enumerate(path):
        if key in tree_loc[loc % 2]:  # noqa: SIM401
            tree_loc[(loc + 1) % 2] = tree_loc[loc % 2][key]
        else:
            tree_loc[loc % 2][key] = tree_loc[(loc + 1) % 2]
        tree_loc[loc % 2] = {}
    if leaves:
        tree_loc[len(path) % 2].update({'leaves': leaves})


def instantiate_middleware(middlewares):
    """Take iterable of middlewares and instantiate.

    Middleware instantiated here will not be shared amongst
    subscriptions/queries.

    """
    for middleware in middlewares:
        if isclass(middleware):
            yield middleware()
            continue
        yield middleware


def null_setter(result):
    """Set type to null if result is empty/null-like."""
    # Only set empty parents to null.
    try:
        if result in EMPTY_VALUES:
            return NULL_VALUE
    except TypeError:
        # If field is a repeated composite field convert to list.
        if not result:
            return NULL_VALUE
    return result


# Is possible to not use middleware and do all the filtering here.
# However, middleware allows for argument of the request doc to set.
def strip_null(data):
    """Recursively strip data structure of nulls."""
    if isinstance(data, Promise):
        return data.then(strip_null)
    if isinstance(data, dict):
        return {
            key: strip_null(val)
            for key, val in data.items()
            if val is not NULL_VALUE
        }
    if isinstance(data, list):
        return [
            strip_null(val)
            for val in data
            if val is not NULL_VALUE
        ]
    return data


def attr_strip_null(result):
    """Work on the attribute/data of ExecutionResult if present."""
    if hasattr(result, 'data'):
        result.data = strip_null(result.data)
        return result
    return strip_null(result)


def null_stripper(exe_result):
    """Strip nulls in accordance with type of execution result."""
    if isinstance(exe_result, Observable):
        return exe_result.map(attr_strip_null)
    if not exe_result.errors:
        return attr_strip_null(exe_result)
    return exe_result


class AstDocArguments:
    """Request doc Argument inspection."""

    def __init__(self, schema, document_ast, variable_values):
        self.schema = schema
        self.operation_defs = {}
        self.fragment_defs = {}
        self.visited_fragments = set()

        for defn in document_ast.definitions:
            if isinstance(defn, ast.OperationDefinition):
                root_type = get_operation_root_type(schema, defn)
                definition_variables = defn.variable_definitions or []
                if definition_variables:
                    def_var_names = {
                        v.variable.name.value
                        for v in definition_variables
                    }
                    var_names_diff = def_var_names.difference({
                        k
                        for k in variable_values
                        if k in def_var_names
                    })
                    # check if we are missing some of the definition variables
                    if var_names_diff:
                        msg = (f'Please check your query variables. The '
                               f'following variables are missing: '
                               f'[{", ".join(var_names_diff)}]')
                        raise ValueError(msg)
                self.operation_defs[getattr(defn.name, 'value', root_type)] = {
                    'definition': defn,
                    'parent_type': root_type,
                    'variables': get_variable_values(
                        schema,
                        definition_variables,
                        variable_values
                    ),
                }
            elif isinstance(defn, ast.FragmentDefinition):
                self.fragment_defs[defn.name.value] = defn

    def has_arg_val(self, arg_name, arg_value):
        """Search through document definitions for argument value.

        Args:
            arg_name (str): Field argument to search for.
            arg_value (Any): Argument value required.

        Returns:

            Boolean

        """
        for components in self.operation_defs.values():
            defn = components['definition']
            if (
                    defn.operation not in STRIP_OPS
                    or getattr(
                        defn.name, 'value', None) == 'IntrospectionQuery'
            ):
                continue
            if self.args_selection_search(
                    components['definition'].selection_set,
                    components['variables'],
                    components['parent_type'],
                    arg_name,
                    arg_value,
            ):
                return True
        return False

    def args_selection_search(
            self, selection_set, variables, parent_type, arg_name, arg_value):
        """Recursively search through feild/fragment selection set fields."""
        for field in selection_set.selections:
            if isinstance(field, ast.FragmentSpread):
                if field.name.value in self.visited_fragments:
                    continue
                frag_def = self.fragment_defs[field.name.value]
                frag_type = type_from_ast(self.schema, frag_def.type_condition)
                if self.args_selection_search(
                        frag_def.selection_set, variables,
                        frag_type, arg_name, arg_value):
                    return True
                self.visited_fragments.add(frag_def.name)
                continue
            field_def = get_field_def(
                self.schema, parent_type, field.name.value)
            if field_def is None:
                continue
            arg_vals = get_argument_values(
                field_def.args, field.arguments, variables)
            if arg_vals.get(arg_name) == arg_value:
                return True
            if field.selection_set is None:
                continue
            if self.args_selection_search(
                    field.selection_set, variables,
                    get_named_type(field_def.type), arg_name, arg_value):
                return True
        return False


def execute_and_validate_and_strip(
    schema: 'GraphQLSchema',
    document_ast: 'Document',
    *args: Any,
    **kwargs: Any
) -> Union['ExecutionResult', Observable]:
    """Wrapper around graphql ``execute_and_validate()`` that adds
    null stripping."""
    result = execute_and_validate(schema, document_ast, *args, **kwargs)
    # Search request document to determine if 'stripNull: true' is set
    # as and argument. It can not be done in the middleware, as they
    # can be Promises/futures (so may not been resolved at this point).
    variable_values = kwargs['variable_values'] or {}
    doc_args = AstDocArguments(schema, document_ast, variable_values)
    if doc_args.has_arg_val(STRIP_ARG, True):
        if kwargs.get('return_promise', False) and hasattr(result, 'then'):
            return result.then(null_stripper)  # type: ignore[union-attr]
        return null_stripper(result)
    return result


class CylcGraphQLBackend(GraphQLBackend):
    """Return a GraphQL document using the default
    graphql executor with optional null-stripping of result.

    The null value stripping of result is triggered by the presence
    of argument & value "stripNull: true" in any field.

    This is a modification of GraphQLCoreBackend found within:
        https://github.com/graphql-python/graphql-core-legacy
    (graphql-core==2.3.2)

    Args:

        executor (object): Executor used in evaluating the resolvers.

    """

    def __init__(self, executor=None):
        self.execute_params = {"executor": executor}

    def document_from_string(self, schema, document_string):
        """Parse string and setup request document for execution.

        Args:

            schema (graphql.GraphQLSchema):
                Schema definition object
            document_string (str):
                Request query/mutation/subscription document.

        Returns:

            graphql.GraphQLDocument

        """
        if isinstance(document_string, ast.Document):
            document_ast = document_string
            document_string = print_ast(document_ast)
        else:
            if not isinstance(document_string, str):
                logger.error("The query must be a string")
            document_ast = parse(document_string)
        return GraphQLDocument(
            schema=schema,
            document_string=document_string,
            document_ast=document_ast,
            execute=partial(
                execute_and_validate_and_strip,
                schema,
                document_ast,
                **self.execute_params
            ),
        )


# -- Middleware --

class IgnoreFieldMiddleware:
    """Set to null/None type undesired field values for stripping."""

    # Sometimes `next` is a Partial(coroutine) or Promise,
    # making inspection for know how to resolve it difficult.
    ASYNC_OPS = {'query', 'mutation'}

    def __init__(self):
        self.args_tree = {}
        self.tree_paths = set()
        self.field_sets = {}

    def resolve(self, next_, root, info, **args):
        """Middleware resolver; handles field according to operation."""
        # GraphiQL introspection is 'query' but not async
        if getattr(info.operation.name, 'value', None) == 'IntrospectionQuery':
            return next_(root, info, **args)

        if info.operation.operation in STRIP_OPS:
            path_string = f'{info.path}'
            # Needed for child fields that resolve without args.
            # Store arguments of parents as leaves of schema tree from path
            # to respective field.
            # no need to regrow the tree on every subscription push/delta
            if args and path_string not in self.tree_paths:
                grow_tree(self.args_tree, info.path, args)
                self.tree_paths.add(path_string)
            if STRIP_ARG not in args:
                branch = self.args_tree
                for section in info.path:
                    if section not in branch:
                        break
                    branch = branch[section]
                    # Only set if present on branch section
                    if 'leaves' in branch and STRIP_ARG in branch['leaves']:
                        args[STRIP_ARG] = branch['leaves'][STRIP_ARG]

            # Now flag empty fields as null for stripping
            if args.get(STRIP_ARG, False):
                field_name = to_snake_case(info.field_name)

                # Clear field set so recreated via first child field,
                # as path may be a parent.
                # Done here as parent may be in NODE_MAP
                if path_string in self.field_sets:
                    del self.field_sets[path_string]

                # Avoid using the protobuf default if field isn't set.
                if (
                    hasattr(root, 'ListFields')
                    and hasattr(root, field_name)
                    and get_named_type(info.return_type).name not in NODE_MAP
                ):

                    # Gather fields set in root
                    parent_path_string = f'{info.path[:-1:]}'
                    stamp = getattr(root, 'stamp', '')
                    if (
                        parent_path_string not in self.field_sets
                        or self.field_sets[
                            parent_path_string]['stamp'] != stamp
                    ):
                        self.field_sets[parent_path_string] = {
                            'stamp': stamp,
                            'fields': {
                                field.name
                                for field, _ in root.ListFields()
                            }
                        }

                    if (
                        parent_path_string in self.field_sets
                        and field_name not in self.field_sets[
                            parent_path_string]['fields']
                    ):
                        return None
                # Do not resolve subfields of an empty type
                # by setting as null in parent/root.
                elif isinstance(root, dict) and field_name in root:
                    field_value = root[field_name]
                    if (
                        field_value in EMPTY_VALUES
                        or (
                            hasattr(field_value, 'ListFields')
                            and not field_value.ListFields()
                        )
                    ):
                        return None
                if (
                    info.operation.operation in self.ASYNC_OPS
                    or iscoroutinefunction(next_)
                ):
                    return self.async_null_setter(next_, root, info, **args)
                return null_setter(next_(root, info, **args))

        if (
            info.operation.operation in self.ASYNC_OPS
            or iscoroutinefunction(next_)
        ):
            return self.async_resolve(next_, root, info, **args)
        return next_(root, info, **args)

    async def async_resolve(self, next_, root, info, **args):
        """Return awaited coroutine"""
        return await next_(root, info, **args)

    async def async_null_setter(self, next_, root, info, **args):
        """Set type to null after awaited result if empty/null-like."""
        result = await next_(root, info, **args)
        return null_setter(result)



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Tasks spawn a sequence of POINTS (P) separated by INTERVALS (I).
Each task may have multiple sequences, e.g. 12-hourly and 6-hourly.
"""

from typing import Optional, Type, overload

from cylc.flow.cycling import PointBase, integer, iso8601
from metomi.isodatetime.data import Calendar


ISO8601_CYCLING_TYPE = iso8601.CYCLER_TYPE_ISO8601
INTEGER_CYCLING_TYPE = integer.CYCLER_TYPE_INTEGER

IS_OFFSET_ABSOLUTE_IMPLS = {
    INTEGER_CYCLING_TYPE: integer.is_offset_absolute,
    ISO8601_CYCLING_TYPE: iso8601.is_offset_absolute,
}


POINTS = {INTEGER_CYCLING_TYPE: integer.IntegerPoint,
          ISO8601_CYCLING_TYPE: iso8601.ISO8601Point}

DUMP_FORMAT_GETTERS = {INTEGER_CYCLING_TYPE: integer.get_dump_format,
                       ISO8601_CYCLING_TYPE: iso8601.get_dump_format}

POINT_RELATIVE_GETTERS = {
    INTEGER_CYCLING_TYPE: integer.get_point_relative,
    ISO8601_CYCLING_TYPE: iso8601.get_point_relative
}

INTERVALS = {INTEGER_CYCLING_TYPE: integer.IntegerInterval,
             ISO8601_CYCLING_TYPE: iso8601.ISO8601Interval}

SEQUENCES = {INTEGER_CYCLING_TYPE: integer.IntegerSequence,
             ISO8601_CYCLING_TYPE: iso8601.ISO8601Sequence}

INIT_FUNCTIONS = {INTEGER_CYCLING_TYPE: integer.init_from_cfg,
                  ISO8601_CYCLING_TYPE: iso8601.init_from_cfg}


class DefaultCycler:

    """Store the default TYPE for Cyclers."""

    TYPE: str


@overload
def get_point(value: str, cycling_type: Optional[str] = None) -> PointBase:
    ...


@overload
def get_point(value: None, cycling_type: Optional[str] = None) -> None:
    ...


def get_point(
    value: Optional[str], cycling_type: Optional[str] = None
) -> Optional[PointBase]:
    """Return a cylc.flow.cycling.PointBase-derived object from a string."""
    if value is None:
        return None
    return get_point_cls(cycling_type=cycling_type)(value)


def get_point_cls(cycling_type: Optional[str] = None) -> Type[PointBase]:
    """Return the cylc.flow.cycling.PointBase-derived class we're using."""
    if cycling_type is None:
        cycling_type = DefaultCycler.TYPE
    return POINTS[cycling_type]


def get_dump_format(cycling_type=None):
    """Return cycle point dump format, or None."""
    return DUMP_FORMAT_GETTERS[cycling_type]()


def get_point_relative(*args, **kwargs):
    """Return a point from an offset expression and a base point."""
    cycling_type = kwargs.pop("cycling_type", DefaultCycler.TYPE)
    return POINT_RELATIVE_GETTERS[cycling_type](*args, **kwargs)


def get_interval(*args, **kwargs):
    """Return a cylc.flow.cycling.IntervalBase-derived object from a string."""
    if args[0] is None:
        return None
    cycling_type = kwargs.pop("cycling_type", DefaultCycler.TYPE)
    return get_interval_cls(cycling_type=cycling_type)(*args, **kwargs)


def get_interval_cls(cycling_type=None):
    """Return the cylc.flow.cycling.IntervalBase-derived class we're using."""
    if cycling_type is None:
        cycling_type = DefaultCycler.TYPE
    return INTERVALS[cycling_type]


def get_sequence(*args, **kwargs):
    """Return a cylc.flow.cycling.SequenceBase-derived object from a string."""
    if args[0] is None:
        return None
    cycling_type = kwargs.pop("cycling_type", DefaultCycler.TYPE)
    return get_sequence_cls(cycling_type=cycling_type)(*args, **kwargs)


def get_sequence_cls(cycling_type=None):
    """Return the cylc.flow.cycling.SequenceBase-derived class we're using."""
    if cycling_type is None:
        cycling_type = DefaultCycler.TYPE
    return SEQUENCES[cycling_type]


def init_cyclers(cfg):
    """Initialise cycling specifics using the workflow configuration (cfg)."""
    DefaultCycler.TYPE = cfg['scheduling']['cycling mode']
    if DefaultCycler.TYPE in Calendar.MODES:
        DefaultCycler.TYPE = ISO8601_CYCLING_TYPE
    INIT_FUNCTIONS[DefaultCycler.TYPE](cfg)


def is_offset_absolute(offset_string, **kwargs):
    """Return True if offset_string is a point rather than an interval."""
    cycling_type = kwargs.pop("cycling_type", DefaultCycler.TYPE)
    return IS_OFFSET_ABSOLUTE_IMPLS[cycling_type](offset_string)


@overload
def standardise_point_string(
    point_string: str, cycling_type: Optional[str] = None
) -> str:
    ...


@overload
def standardise_point_string(
    point_string: None, cycling_type: Optional[str] = None
) -> None:
    ...


def standardise_point_string(
    point_string: Optional[str], cycling_type: Optional[str] = None
) -> Optional[str]:
    """Return a standardised version of point_string."""
    if point_string is None:
        return None
    point = get_point(point_string, cycling_type=cycling_type)
    if point is not None:
        point.standardise()
        point_string = str(point)
    return point_string



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""This module is for parsing date/time representations.

Expressions can be formatted in either the full ISO 8601:2004 date/time
representations (also including the ISO 8601:2000 truncated syntax), or
in the Cylc-specific abbreviations thereof.

Where abbreviations or truncations are used, the missing information
needs to be filled in by some date/time context - this means that "T06"
needs to be applied to a context time such as "20201205T0000Z" in order
to get a full date/time (in this example, "20201205T0600Z"). This
context is the initial cycle point or final cycle point for the task
cycling dependency graph section specification, and is the task cycle
time for inter-cycle task references such as "foo[-P6Y] => foo".

"""

import re
from typing import TYPE_CHECKING, List, Optional, Pattern, Tuple, Union, cast

from metomi.isodatetime.data import Duration, TimeRecurrence
from metomi.isodatetime.exceptions import IsodatetimeError
from metomi.isodatetime.parsers import (
    DurationParser, TimePointParser, TimeRecurrenceParser
)

from cylc.flow import LOG
from cylc.flow.cycling import parse_exclusion
from cylc.flow.exceptions import (
    CylcMissingContextPointError,
    CylcMissingFinalCyclePointError,
    CylcTimeSyntaxError,
)
import cylc.flow.flags

if TYPE_CHECKING:
    from metomi.isodatetime.data import TimePoint
    from cylc.flow.cycling.iso8601 import ISO8601Point


UTC_UTC_OFFSET_HOURS_MINUTES = (0, 0)


class CylcTimeParser:

    """Parser for Cylc abbreviated/full ISO 8601 syntax.

    Arguments:
    context_start_point describes the beginning date/time used for
    extrapolating incomplete date/time syntax (usually initial
    cycle point). It is either a date/time string in full ISO 8601
    syntax or a metomi.isodatetime.data.TimePoint object.
    context_end_point is the same as context_start_point, but describes
    a final time - e.g. the final cycle point.

    """

    POINT_INVALID_FOR_CYLC_REGEXES = [
        (re.compile(r"^\d\d$"),
         ("2 digit centuries not allowed. Did you mean T-digit-digit e.g. "
          "'T00'?"))
    ]

    RECURRENCE_FORMAT_REGEXES: List[Tuple[Pattern, int]] = [
        (re.compile(r"^(?P<start>[^PR/][^/]*)$"), 3),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<start>[^PR/][^/]*)/(?P<end>[^PR/]"
                    "[^/]*)$"), 1),
        (re.compile(r"^(?P<start>[^PR/][^/]*)/(?P<intv>P[^/]*)/?$"), 3),
        (re.compile(r"^(?P<intv>P[^/]*)$"), 3),
        (re.compile(r"^(?P<intv>P[^/]*)/(?P<end>[^PR/][^/]*)$"), 4),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<start>[^PR/][^/]*)/?$"), 3),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<start>[^PR/][^/]*)/(?P<intv>P[^/]"
                    "*)$"), 3),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<start>)/(?P<intv>P[^/]*)$"), 3),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<intv>P[^/]*)/(?P<end>[^PR/][^/]*)"
                    "$"), 4),
        (re.compile(r"^R(?P<reps>\d+)?/(?P<intv>P[^/]*)/?$"), 4),
        (re.compile(r"^R(?P<reps>\d+)?//(?P<end>[^PR/][^/]*)$"), 4),
        (re.compile(r"^R(?P<reps>1)/?(?P<start>$)"), 3),
        (re.compile(r"^R(?P<reps>1)//(?P<end>[^PR/][^/]*)$"), 4)
    ]

    CHAIN_REGEX = re.compile(r'((?:[+-P]|[\dT])[\d\w]*)')

    MIN_REGEX = re.compile(r'min\(([^)]+)\)')

    OFFSET_REGEX = re.compile(r"(?P<sign>[+-])(?P<intv>P.+)$")

    TRUNCATED_REC_MAP = {
        "---": [re.compile(r"^\d\dT")],
        "--": [re.compile(r"^\d\d\d\dT")],
        "-": [
            re.compile(r"^\d\d\dT"),
            re.compile(r"\dW\d\dT"),
            re.compile(r"W\d\d\d?T"),
            re.compile(r"W-\dT"),
            re.compile(r"W-\d")
        ]
    }

    __slots__ = ('timepoint_parser', 'duration_parser', 'recurrence_parser',
                 'context_start_point', 'context_end_point')

    def __init__(
        self,
        context_start_point: Union['ISO8601Point', 'TimePoint', str, None],
        context_end_point: Union['ISO8601Point', 'TimePoint', str, None],
        parsers: Tuple[TimePointParser, DurationParser, TimeRecurrenceParser]
    ):
        if context_start_point is not None:
            context_start_point = str(context_start_point)
        if context_end_point is not None:
            context_end_point = str(context_end_point)
        self.timepoint_parser, self.duration_parser, self.recurrence_parser = (
            parsers
        )

        if isinstance(context_start_point, str):
            context_start_point, _ = self._get_point_from_expression(
                context_start_point, None)
        self.context_start_point: Optional['TimePoint'] = context_start_point
        if isinstance(context_end_point, str):
            context_end_point, _ = self._get_point_from_expression(
                context_end_point, None)
        self.context_end_point: Optional['TimePoint'] = context_end_point

    @staticmethod
    def initiate_parsers(num_expanded_year_digits=0, dump_format=None,
                         assumed_time_zone=None):
        """Initiate datetime parsers required to initiate this class.

        Returns:
            tuple: (timepoint_parser, duration_parser, time_recurrence_parser)
                - timepoint_parser - metomi.isodatetime.parsers.TimePointParser
                  obj.
                - duration_parser - metomi.isodatetime.parsers.DurationParser
                  obj.
                - time_recurrence_parser -
                  metomi.isodatetime.parsers.TimeRecurrenceParser obj
        """

        if dump_format is None:
            if num_expanded_year_digits:
                dump_format = "+XCCYYMMDDThhmmZ"
            else:
                dump_format = "CCYYMMDDThhmmZ"

        timepoint_parser = TimePointParser(
            allow_only_basic=False,
            allow_truncated=True,
            num_expanded_year_digits=num_expanded_year_digits,
            dump_format=dump_format,
            assumed_time_zone=assumed_time_zone
        )

        return (timepoint_parser, DurationParser(), TimeRecurrenceParser())

    def parse_interval(self, expr: str) -> Duration:
        """Parse an interval (duration) in full ISO date/time format."""
        return self.duration_parser.parse(expr)

    def parse_timepoint(
        self, expr: str, context_point: Optional['TimePoint'] = None
    ) -> 'TimePoint':
        """Parse an expression in abbrev. or full ISO date/time format.

        Args:
            expr: a string such as 20010205T00Z, or a truncated/abbreviated
                format string such as T06 or -P2Y.
            context_point: supplies the missing information for truncated
                expressions. E.g. for inter-cycle dependency expressions,
                it should be the task cycle point.
                If None, self.context_start_point is used.

        """
        if context_point is None:
            context_point = self.context_start_point
        point, offsets = self._get_point_from_expression(
            expr, context_point, allow_truncated=True
        )
        if point is not None:
            if point.truncated:
                point += context_point  # type: ignore[operator]
            for offset in offsets:
                point += offset
            return point
        raise CylcTimeSyntaxError(
            ("'%s': not a valid cylc-shorthand or full " % expr) +
            "ISO 8601 date representation")

    def parse_recurrence(
        self,
        expression: str,
        context_start_point: Optional['TimePoint'] = None,
        context_end_point: Optional['TimePoint'] = None,
        zero_duration_warning: bool = True,
    ) -> Tuple[TimeRecurrence, list]:
        """Parse an expression in abbrev. or full ISO recurrence format.

        Args:
            expression:
                The recurrence expression to parse.
            context_start_point:
                Sequence start point from the global context.
            context_end_point:
                Sequence end point from the global context.
            zero_duration_warning:
                If `False`, then zero-duration recurrence warnings will be
                turned off. This is set for exclusion parsing.

        """
        expression, exclusions = parse_exclusion(str(expression))

        if context_start_point is None:
            context_start_point = self.context_start_point
        if context_end_point is None:
            context_end_point = self.context_end_point
        for rec_object, format_num in self.RECURRENCE_FORMAT_REGEXES:
            result = rec_object.search(expression)
            if not result:
                continue

            repetitions = result.groupdict().get("reps")
            if repetitions is not None:
                repetitions = int(repetitions)
            start: Optional[str] = result.groupdict().get("start")
            end: Optional[str] = result.groupdict().get("end")
            start_required = (format_num in {1, 3})
            end_required = (format_num in {1, 4})
            start_point, start_offsets = self._get_point_from_expression(
                start, context_start_point,
                is_required=start_required,
                allow_truncated=True
            )
            try:
                end_point, end_offsets = self._get_point_from_expression(
                    end, context_end_point,
                    is_required=end_required,
                    allow_truncated=True
                )
            except CylcMissingContextPointError:
                raise CylcMissingFinalCyclePointError(
                    "This workflow requires a final cycle point."
                )

            exclusion_points = []
            # Convert the exclusion strings to ISO8601 points
            if exclusions is not None:
                for exclusion in exclusions:
                    try:
                        # Attempt to convert to TimePoint
                        exclusion_point, excl_offsets = (
                            self._get_point_from_expression(exclusion, None)
                        )
                        for offset in excl_offsets:
                            exclusion_point += offset  # type: ignore[operator]
                        exclusion_points.append(exclusion_point)
                    except (CylcTimeSyntaxError, IndexError):
                        # Not a point, parse it as recurrence later
                        exclusion_points.append(exclusion)

            intv = result.groupdict().get("intv")
            intv_context_truncated_point = None
            if start_point is not None and start_point.truncated:
                intv_context_truncated_point = start_point
            if end_point is not None and end_point.truncated:
                intv_context_truncated_point = end_point
            interval = self._get_interval_from_expression(
                intv, context=intv_context_truncated_point)
            if format_num == 1:
                interval = None
            elif repetitions == 1:
                interval = Duration(0)
            if start_point is not None:
                if start_point.truncated:
                    start_point += (  # type: ignore[operator]
                        context_start_point
                    )
                for offset in start_offsets:
                    start_point += offset
            if end_point is not None:
                if end_point.truncated:
                    end_point += context_end_point  # type: ignore[operator]
                for offset in end_offsets:
                    end_point += offset

            if (
                interval and
                start_point is None and repetitions is None and
                context_start_point is not None
            ):
                # isodatetime only reverses bounded end-point recurrences.
                # This is unbounded, and will come back in reverse order.
                # We need to reverse it.
                start_point = cast(  # (end pt can't be None if start is None)
                    'TimePoint', end_point
                )
                repetitions = 1
                while start_point > context_start_point:
                    start_point -= interval
                    repetitions += 1
                end_point = None

            if (
                zero_duration_warning
                and not interval
                and repetitions != 1
                and (format_num != 1 or start_point == end_point)
            ):
                LOG.warning(
                    "Cannot have more than 1 repetition for zero-duration "
                    f"recurrence {expression}."
                )

            if cylc.flow.flags.cylc7_back_compat and format_num == 1:
                LOG.warning(
                    f"The recurrence '{expression}' is unlikely to behave "
                    "the same way as in Cylc 7 as that implementation was "
                    "incorrect (see https://cylc.github.io/cylc-doc/stable/"
                    "html/user-guide/writing-workflows/scheduling.html"
                    "#format-1-r-limit-datetime-datetime)"
                )

            return TimeRecurrence(
                repetitions=repetitions,
                start_point=start_point,
                duration=interval,
                end_point=end_point
            ), exclusion_points

        raise CylcTimeSyntaxError("Could not parse %s" % expression)

    def _get_interval_from_expression(
        self, expr: Optional[str], context: Optional['TimePoint'] = None
    ) -> Optional[Duration]:
        if expr is None:
            if context is None or not context.truncated:
                return None
            prop_name = context.get_largest_truncated_property_name()
            kwargs = {}
            if prop_name == "year_of_century":
                kwargs = {"years": 100}
            if prop_name == "year_of_decade":
                kwargs = {"years": 10}
            if prop_name in ["month_of_year", "week_of_year", "day_of_year"]:
                kwargs = {"years": 1}
            if prop_name == "day_of_month":
                kwargs = {"months": 1}
            if prop_name == "day_of_week":
                kwargs = {"days": 7}
            if prop_name == "hour_of_day":
                kwargs = {"days": 1}
            if prop_name == "minute_of_hour":
                kwargs = {"hours": 1}
            if prop_name == "second_of_minute":
                kwargs = {"minutes": 1}
            if not kwargs:
                return None
            return Duration(**kwargs)
        return self.duration_parser.parse(expr)

    def _get_min_from_expression(
        self,
        expr: str,
        context: Optional['TimePoint']
    ) -> str:
        points: List[str] = [
            x.strip() for x in re.findall(self.MIN_REGEX, expr)[0].split(",")
        ]
        ptslist: List['TimePoint'] = []
        min_entry = ""
        for point in points:
            cpoint, offsets = self._get_point_from_expression(
                point, context, allow_truncated=True
            )
            if cpoint is not None:
                if cpoint.truncated:
                    cpoint += context  # type: ignore[operator]
                for offset in offsets:
                    cpoint += offset
                ptslist.append(cpoint)
                if cpoint == min(ptslist):
                    min_entry = point
        return min_entry

    def _get_point_from_expression(
        self,
        expr: Optional[str],
        context: Optional['TimePoint'],
        is_required: bool = False,
        allow_truncated: bool = False
    ) -> Tuple[Optional['TimePoint'], List[Duration]]:
        """Gets a TimePoint from an expression"""
        if expr is None:
            if is_required and allow_truncated:
                if context is None:
                    raise CylcMissingContextPointError(
                        "Missing context cycle point."
                    )
                return context, []
            return None, []

        if expr.startswith("min("):
            expr = self._get_min_from_expression(expr, context)

        expr_offsets: List[Duration] = []
        if self.OFFSET_REGEX.search(expr):
            expr, expr_offsets = self.parse_chain_expression(expr)
        if not expr and allow_truncated:
            return context, expr_offsets

        for invalid_rec, msg in self.POINT_INVALID_FOR_CYLC_REGEXES:
            if invalid_rec.search(expr):
                raise CylcTimeSyntaxError(f"'{expr}': {msg}")

        expr_to_parse = expr
        if expr.endswith("T"):
            expr_to_parse += "00"
        try:
            expr_point: 'TimePoint' = self.timepoint_parser.parse(
                expr_to_parse
            )
        except ValueError:  # not IsodatetimeError as too specific
            pass
        else:
            return expr_point, expr_offsets

        if allow_truncated:
            for truncation_string, recs in self.TRUNCATED_REC_MAP.items():
                for rec in recs:
                    if rec.search(expr):
                        try:
                            expr_point = self.timepoint_parser.parse(
                                truncation_string + expr_to_parse)
                        except IsodatetimeError:
                            continue
                        return expr_point, expr_offsets

        raise CylcTimeSyntaxError(
            f"'{expr}': not a valid cylc-shorthand or full "
            "ISO 8601 date representation"
        )

    def parse_chain_expression(self, expr: str) -> Tuple[str, List[Duration]]:
        """Parse an expression such as '+P1M+P1D'.

        Returns:
            Expression: The expression with any offsets removed.
            Offsets: List of offsets from the expression. Note: by keeping
                offsets separate (rather than combine into 1 Duration),
                we preserve order of operations.

        Examples:
            >>> ctp = CylcTimeParser(
            ...    None, None, (TimePointParser(), DurationParser(), None)
            ... )
            >>> expr, offsets = ctp.parse_chain_expression('2022+P1M-P1D')
            >>> expr
            '2022'
            >>> [str(i) for i in offsets]
            ['P1M', '-P1D']
        """
        expr_offsets: List[Duration] = []
        chain_expr: List[str] = self.CHAIN_REGEX.findall(expr)
        expr = ""
        for item in chain_expr:
            if "P" not in item:
                expr += item
                continue
            split_expr = self.OFFSET_REGEX.split(item)
            expr += split_expr.pop(0)
            expr_offset_item = self.duration_parser.parse(item[1:])
            if item[0] == "-":
                expr_offset_item *= -1
            expr_offsets.append(expr_offset_item)
        return expr, expr_offsets



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Functions relating to (job) platforms."""

import random
import re
from copy import deepcopy
from typing import (
    TYPE_CHECKING, Any, Dict, Iterable,
    List, Optional, Set, Union, overload
)

from cylc.flow import LOG

from cylc.flow.exceptions import (
    GlobalConfigError,
    PlatformLookupError, CylcError, NoHostsError, NoPlatformsError)
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
from cylc.flow.hostuserutil import is_remote_host

if TYPE_CHECKING:
    from cylc.flow.parsec.OrderedDict import OrderedDictWithDefaults

UNKNOWN_TASK = 'unknown task'

FORBIDDEN_WITH_PLATFORM: Dict[str, Dict[str, Set[Optional[str]]]] = {
    'remote': {
        # setting: exclusions
        'host': {'localhost', None},
        'retrieve job logs': {None},
        'retrieve job logs max size': {None},
        'retrieve job logs retry delays': {None},
    },
    'job': {
        'batch system': {None},
        'batch submit command template': {None}
    }
}

DEFAULT_JOB_RUNNER = 'background'
SINGLE_HOST_JOB_RUNNERS = ['background', 'at']

# Regex to check whether a string is a command
HOST_REC_COMMAND = re.compile(r'(`|\$\()\s*(.*)\s*([`)])$')
PLATFORM_REC_COMMAND = re.compile(r'(\$\()\s*(.*)\s*([)])$')

HOST_SELECTION_METHODS = {
    'definition order': lambda goodhosts: goodhosts[0],
    'random': random.choice
}


def log_platform_event(
    event: str,
    platform: dict,
    host: Optional[str] = None,
    level: str = 'info'
):
    """Log a simple platform event."""
    # matches cylc.flow.exceptions.PlatformError format
    getattr(LOG, level)(
        f'platform: {platform["name"]} - {event}'
        + (f' (on {host})' if host else '')
    )


@overload
def get_platform(
    task_conf: Optional[str] = None,
    task_name: str = UNKNOWN_TASK,
    bad_hosts: Optional[Set[str]] = None
) -> Dict[str, Any]:
    ...


@overload
def get_platform(
    task_conf: Union[dict, 'OrderedDictWithDefaults'],
    task_name: str = UNKNOWN_TASK,
    bad_hosts: Optional[Set[str]] = None
) -> Optional[Dict[str, Any]]:
    ...


# BACK COMPAT: get_platform
#     At Cylc 8.x remove all Cylc7 upgrade logic.
# from:
#     Cylc8
# to:
#     Cylc8.x
# remove at:
#     Cylc8.x
def get_platform(
    task_conf: Union[str, dict, 'OrderedDictWithDefaults', None] = None,
    task_name: str = UNKNOWN_TASK,
    bad_hosts: Optional[Set[str]] = None
) -> Optional[Dict[str, Any]]:
    """Get a platform.

    Looking at a task config this method decides whether to get platform from
    name, or Cylc7 config items.

    Args:
        task_conf: If str this is assumed to be the platform name, otherwise
            this should be a configuration for a task.
        task_name: Help produce more helpful error messages.
        bad_hosts: A set of hosts known to be unreachable (had an ssh 255
            error)

    Returns:
        platform: A platform definition dictionary. Uses either
            get_platform() or platform_name_from_job_info(), but to the
            user these look the same.

    Raises:
        NoPlatformsError:
            Platform group has no platforms with usable hosts.
            This should be caught if this function is used on a raw config,
            or in any other context where a platform group might be selected.
        PlatformLookupError:
            Raised if the name of a platform cannot be selected based on the
            information given.
    """
    if task_conf is None or isinstance(task_conf, str):  # noqa: SIM 114
        # task_conf is a platform name, or get localhost if None
        return platform_from_name(task_conf, bad_hosts=bad_hosts)

    # NOTE: Do NOT use .get() on OrderedDictWithDefaults -
    # https://github.com/cylc/cylc-flow/pull/4975
    elif 'platform' in task_conf and task_conf['platform']:
        # Check whether task has conflicting Cylc7 items.
        fail_if_platform_and_host_conflict(task_conf, task_name)

        if is_platform_definition_subshell(task_conf['platform']):
            # Platform definition is using subshell e.g. platform = $(foo);
            # won't be evaluated until job submit so cannot get or
            # validate platform
            return None

        # If platform name exists and doesn't clash with Cylc7 Config items.
        return platform_from_name(task_conf['platform'], bad_hosts=bad_hosts)

    else:
        if get_platform_deprecated_settings(task_conf) == []:
            # No deprecated items; platform is localhost
            return platform_from_name()
        else:
            # Need to calculate platform
            # NOTE: Do NOT use .get() on OrderedDictWithDefaults - see above
            task_job_section = task_conf['job'] if 'job' in task_conf else {}
            task_remote_section = (
                task_conf['remote'] if 'remote' in task_conf else {})
            return platform_from_name(
                platform_name_from_job_info(
                    glbl_cfg().get(['platforms']),
                    task_job_section,
                    task_remote_section
                ),
                bad_hosts=bad_hosts
            )


def platform_from_name(
    platform_name: Optional[str] = None,
    platforms: Optional[Dict[str, Dict[str, Any]]] = None,
    bad_hosts: Optional[Set[str]] = None
) -> Dict[str, Any]:
    """
    Find out which job platform to use given a list of possible platforms and
    a task platform string.

    Verifies selected platform is present in global.cylc file and returns it,
    raises error if platform is not in global.cylc or returns 'localhost' if
    no platform is initially selected.

    Args:
        platform_name: name of platform to be retrieved.
        platforms: global.cylc platforms given as a dict.
        bad_hosts: A set of hosts known to be unreachable (had an ssh 255
            error)

    Returns:
        platform: object containing settings for a platform, loaded from
            Global Config.

    Raises:
        NoPlatformsError: Platform group has no platforms with usable hosts.
    """
    if platforms is None:
        platforms = glbl_cfg().get(['platforms'])
    platform_groups = glbl_cfg().get(['platform groups'])

    if platform_name is None:
        platform_name = 'localhost'

    # The list is reversed to allow user-set platform groups (which are
    # appended to site set platform groups) to be matched first and override
    # site defined platform groups.
    for platform_name_re in reversed(list(platform_groups)):
        # Platform is member of a group.
        if re.fullmatch(platform_name_re, platform_name):
            platform_name = get_platform_from_group(
                platform_groups[platform_name_re], group_name=platform_name,
                bad_hosts=bad_hosts
            )
            break

    for platform_name_re in platforms:
        if (
            # If the platform_name_re contains special regex chars
            re.escape(platform_name_re) != platform_name_re
            and re.match(platform_name_re, 'localhost')
        ):
            raise PlatformLookupError(
                'The "localhost" platform cannot be defined using a '
                'regular expression. See the documentation for '
                '"global.cylc[platforms][localhost]" for more information.'
            )

    # The list is reversed to allow user-set platforms (which are appended to
    # than site set platforms) to be matched first and override site defined
    # platforms.
    for platform_name_re in reversed(list(platforms)):
        # We substitute commas with or without spaces to
        # allow lists of platforms
        if re.fullmatch(
            re.sub(
                r'\s*(?!{[\s\d]*),(?![\s\d]*})\s*',
                '|',
                platform_name_re
            ),
            platform_name
        ):
            # Deepcopy prevents contaminating platforms with data
            # from other platforms matching platform_name_re
            platform_data = deepcopy(platforms[platform_name_re])

            # If hosts are not filled in make remote
            # hosts the platform name.
            # Example: `[platforms][workplace_vm_123]<nothing>`
            #   should create a platform where
            #   `hosts = ['workplace_vm_123']`
            # NOTE: Probably don't use .get() due to OrderedDictWithDefaults -
            # see https://github.com/cylc/cylc-flow/pull/4975
            if (
                'hosts' not in platform_data or
                not platform_data['hosts']
            ):
                platform_data['hosts'] = [platform_name]
            # Fill in the "private" name field.
            platform_data['name'] = platform_name
            return platform_data

    raise PlatformLookupError(
        f"No matching platform \"{platform_name}\" found")


def get_platform_from_group(
    group: Union[dict, 'OrderedDictWithDefaults'],
    group_name: str,
    bad_hosts: Optional[Set[str]] = None
) -> str:
    """Get platform name from group, according to the selection method.

    Args:
        group: A platform group dictionary.
        group_name: Name of the group.
        bad_hosts: The set of hosts found to be unreachable.

    Returns:
        Name of platform selected, or False if all hosts on all platforms are
        in bad_hosts.

    Raises:
        NoPlatformsError: If there are no platforms with any usable
        hosts in the platform group.

    TODO: Uses host_selection methods; should also allow custom select methods.
    """
    if bad_hosts:
        platform_names = [
            platform for platform in group['platforms']
            if any(
                host not in bad_hosts
                for host in platform_from_name(platform)['hosts']
            )
        ]
    else:
        platform_names = group['platforms']

    # If there are no platforms available to be selected:
    if not platform_names:
        hosts_consumed = {
            host
            for platform in group['platforms']
            for host in platform_from_name(platform)['hosts']}
        raise NoPlatformsError(
            group_name, hosts_consumed)

    # Get the selection method
    method = group['selection']['method']
    if method not in HOST_SELECTION_METHODS:
        raise CylcError(
            f'\"{method}\" is not a supported platform selection method.'
        )
    else:
        return HOST_SELECTION_METHODS[method](platform_names)


def platform_name_from_job_info(
    platforms: Union[dict, 'OrderedDictWithDefaults'],
    job: Dict[str, Any],
    remote: Dict[str, Any]
) -> str:
    """
    Find out which job platform to use given a list of possible platforms
    and the task dictionary with cylc 7 definitions in it.

    (Note: "batch system" (Cylc 7) and "job runner" (Cylc 8)
    mean the same thing)

          +------------+ Yes    +-----------------------+
    +-----> Tried all  +------->+ RAISE                 |
    |     | platforms? |        | PlatformNotFoundError |
    |     +------------+        +-----------------------+
    |              No|
    |     +----------v---+
    |     | Examine next |
    |     | platform     |
    |     +--------------+
    |                |
    |     +----------v----------------+
    |     | Do all items other than   |
    +<----+ "host" and "batch system" |
    |   No| match for this plaform    |
    |     +---------------------------+
    |                           |Yes
    |                +----------v----------------+ No
    |                | Task host is 'localhost'? +--+
    |                +---------------------------+  |
    |                           |Yes                |
    |              No+----------v----------------+  |
    |            +---+ Task batch system is      |  |
    |            |   | 'background'?             |  |
    |            |   +---------------------------+  |
    |            |              |Yes                |
    |            |   +----------v----------------+  |
    |            |   | RETURN 'localhost'        |  |
    |            |   +---------------------------+  |
    |            |                                  |
    |    +-------v-------------+     +--------------v-------+
    |  No| batch systems match |  Yes| batch system and     |
    +<---+ and 'localhost' in  |  +--+ host both match      |
    |    | platform hosts?     |  |  +----------------------+
         +---------------------+  |                 |No
    |            |Yes             |  +--------------v-------+
    |    +-------v--------------+ |  | batch system match   |
    |    | RETURN this platform <-+--+ and regex of platform|
    |    +----------------------+ Yes| name matches host    |
    |                                +----------------------+
    |                                  |No
    +<---------------------------------+

    Args:
        job: Workflow config [runtime][TASK][job] section.
        remote: Workflow config [runtime][TASK][remote] section.
        platforms: Dictionary containing platform definitions.

    Returns:
        platform: string representing a platform from the global config.

    Raises:
        PlatformLookupError:
            If no matching platform can be a found an error is raised.

    Example:
        >>> platforms = {
        ...         'desktop[0-9][0-9]|laptop[0-9][0-9]': {},
        ...         'sugar': {
        ...             'hosts': 'localhost',
        ...             'job runner': 'slurm'
        ...         }
        ... }
        >>> job = {'batch system': 'slurm'}
        >>> remote = {'host': 'localhost'}
        >>> platform_name_from_job_info(platforms, job, remote)
        'sugar'
        >>> remote = {}
        >>> platform_name_from_job_info(platforms, job, remote)
        'sugar'
        >>> remote ={'host': 'desktop92'}
        >>> job = {}
        >>> platform_name_from_job_info(platforms, job, remote)
        'desktop92'
    """

    # These settings are removed from the incoming dictionaries for special
    # handling later - we want more than a simple match:
    #   - In the case of "host" we also want a regex match to the platform name
    #   - In the case of "batch system" we want to match the name of the
    #     system/job runner to a platform when host is localhost.

    # NOTE: Do NOT use .get() on OrderedDictWithDefaults -
    # https://github.com/cylc/cylc-flow/pull/4975
    if 'host' in remote and remote['host']:
        task_host = remote['host']
    else:
        task_host = 'localhost'
    if 'batch system' in job and job['batch system']:
        task_job_runner = job['batch system']
    else:
        # Necessary? Perhaps not if batch system default is 'background'
        task_job_runner = 'background'

    # Riffle through the platforms looking for a match to our task settings.
    # reverse dict order so that user config platforms added last are examined
    # before site config platforms.
    for platform_name, platform_spec in reversed(list(platforms.items())):
        # Handle all the items requiring an exact match.
        # All items other than batch system and host must be an exact match
        if not generic_items_match(platform_spec, job, remote):
            continue
        # We have some special logic to identify whether task host and task
        # batch system match the platform in question.
        if (
                not is_remote_host(task_host) and
                task_job_runner == 'background'
        ):
            return 'localhost'

        elif (
            'hosts' in platform_spec and
            task_host in platform_spec['hosts'] and
            task_job_runner == platform_spec['job runner']
        ):
            # If we have localhost with a non-background batch system we
            # use the batch system to give a sensible guess at the platform
            return platform_name

        elif (
            re.fullmatch(platform_name, task_host) and (
                (
                    task_job_runner == 'background' and
                    'job runner' not in platform_spec
                ) or
                task_job_runner == platform_spec['job runner']
            )
        ):
            return task_host

    raise PlatformLookupError('No platform found matching your task')


def generic_items_match(
    platform_spec: Dict[str, Any],
    job: Dict[str, Any],
    remote: Dict[str, Any]
) -> bool:
    """Checks generic items from job/remote against a platform.

    We carry out extra checks on ``[remote]host`` and ``[job]batch system``
    but all other set config items must match between a platform and old
    settings for that platform to match the legacy settings.

    Args:
        platform_spec: Dictionary of platform spec.
        job: Dictionary of config spec section ``[runtime][TASK][job]``
        remote: Dictionary of config spec section ``[runtime][TASK][remote]``

    Returns:
        Does this platform have generic items (not ``[job]batch system``
        or ``[remote]host`` which are treated specially) that are the same.
    """
    # Don't check Host and batch system - they have their own logic.
    if 'host' in remote:
        remote_generic = {
            k: v for k, v in remote.items()
            if k != 'host' and v is not None
        }
    else:
        remote_generic = remote
    if 'batch system' in job:
        job_generic = {
            k: v for k, v in job.items()
            if k != 'batch system' and v is not None
        }
    else:
        job_generic = job

    for task_section in [job_generic, remote_generic]:
        # Get a set of items actually set in both platform and task_section.
        shared_items = set(task_section).intersection(set(platform_spec))
        # If any set items do not match, we can't use this platform.
        if not all(
            platform_spec[item] == task_section[item]
            for item in shared_items
        ):
            return False
    return True


def get_host_from_platform(
    platform: Dict[str, Any], bad_hosts: Optional[Set[str]] = None
) -> str:
    """Placeholder for a more sophisticated function which returns a host
    given a platform dictionary.

    Args:
        platform: A dict representing a platform.
        bad_hosts: A set of hosts Cylc knows to be unreachable.

    Returns:
        hostname: The name of a host.

    Raises:
        NoHostsError:
            This error should be caught by caller to prevent workflow shutdown.
    """
    # Get list of goodhosts:
    if bad_hosts:
        goodhosts = [i for i in platform['hosts'] if i not in bad_hosts]
    else:
        goodhosts = platform['hosts']

    # Get the selection method
    method = platform['selection']['method']
    if not goodhosts:
        raise NoHostsError(platform)
    else:
        if method not in HOST_SELECTION_METHODS:
            raise CylcError(
                f'method \"{method}\" is not a supported host '
                'selection method.'
            )
        else:
            return HOST_SELECTION_METHODS[method](goodhosts)


def fail_if_platform_and_host_conflict(
    task_conf: Union[dict, 'OrderedDictWithDefaults'], task_name: str
) -> None:
    """Raise an error if [runtime][<task>] spec contains platform and
    forbidden host items.

    Args:
        task_conf: A specification to be checked.
        task_name: A name to be given in an error.

    Raises:
        PlatformLookupError - if platform and host items conflict

    """
    # NOTE: Do NOT use .get() on OrderedDictWithDefaults -
    # https://github.com/cylc/cylc-flow/pull/4975
    if 'platform' in task_conf and task_conf['platform']:
        fail_items = [
            f'\n * [{section}]{key}'
            for section, keys in FORBIDDEN_WITH_PLATFORM.items()
            if section in task_conf
            for key, _ in keys.items()
            if (
                key in task_conf[section] and
                task_conf[section][key] is not None
            )
        ]
        if fail_items:
            raise PlatformLookupError(
                f"Task '{task_name}' has the following deprecated '[runtime]' "
                "setting(s) which cannot be used with "
                f"'platform = {task_conf['platform']}':{''.join(fail_items)}"
            )


def get_platform_deprecated_settings(
    task_conf: Union[dict, 'OrderedDictWithDefaults'],
    task_name: str = UNKNOWN_TASK
) -> List[str]:
    """Return deprecated [runtime][<task_name>] settings that should be
    upgraded to platforms.

    Args:
        task_conf: Runtime configuration for the task.
        task_name: The task name.
    """
    return [
        f'[runtime][{task_name}][{section}]{key} = {task_conf[section][key]}'
        for section, keys in FORBIDDEN_WITH_PLATFORM.items()
        if section in task_conf
        for key, exclusions in keys.items()
        if (
            key in task_conf[section] and
            task_conf[section][key] not in exclusions
        )
    ]


def is_platform_definition_subshell(value: str) -> bool:
    """Is the platform definition using subshell? E.g. platform = $(foo)

    Raise PlatformLookupError if using backticks.
    """
    if PLATFORM_REC_COMMAND.match(value):
        return True
    if HOST_REC_COMMAND.match(value):
        raise PlatformLookupError(
            f"platform = {value}: backticks are not supported; please use $()"
        )
    return False


def get_install_target_from_platform(platform: Dict[str, Any]) -> str:
    """Sets install target to configured or default platform name.

    Returns install target.
    """
    if not platform['install target']:
        platform['install target'] = platform['name']

    return platform['install target']


def get_install_target_to_platforms_map(
    platform_names: Iterable[str],
    quiet: bool = False
) -> Dict[str, List[Dict[str, Any]]]:
    """Get a dictionary of unique install targets and the platforms which use
    them.

    Args:
        platform_names: List of platform names to look up in the global config.
        quiet: Supress PlatformLookupErrors

    Return {install_target_1: [platform_1_dict, platform_2_dict, ...], ...}
    """
    ret: Dict[str, List[Dict[str, Any]]] = {}
    for p_name in set(platform_names):
        try:
            platform = platform_from_name(p_name)
        except PlatformLookupError as exc:
            if not quiet:
                raise exc
        else:
            install_target = get_install_target_from_platform(platform)
            ret.setdefault(install_target, []).append(platform)
    return ret


def is_platform_with_target_in_list(
        install_target: str,
        distinct_platforms_list: Iterable[Dict[str, Any]]
) -> bool:
    """Determines whether install target is in the list of platforms"""
    return any(
        install_target == distinct_platform['install target']
        for distinct_platform in distinct_platforms_list
    )


def get_localhost_install_target() -> str:
    """Returns the install target of localhost platform"""
    localhost = get_platform()
    return get_install_target_from_platform(localhost)


def _validate_single_host(
    platforms_cfg: Union[dict, 'OrderedDictWithDefaults']
) -> None:
    """Check that single-host platforms only specify a single host.

    Some job runners don't work across multiple hosts; the job ID is only valid
    on the specific submission host.
    """
    bad_platforms = []
    runners = set()
    name: str
    config: dict
    for name, config in platforms_cfg.items():
        runner = config.get('job runner', DEFAULT_JOB_RUNNER)
        hosts = config.get('hosts', [])
        if runner in SINGLE_HOST_JOB_RUNNERS and len(hosts) > 1:
            bad_platforms.append((name, runner, hosts))
            runners.add(runner)
    if bad_platforms:
        if len(runners) > 1:
            grammar = ["are", "s"]
        else:
            grammar = ["is a", ""]
        msg = (
            f"{', '.join(runners)} {grammar[0]} single-host"
            f" job runner{grammar[1]}:"
        )
        for name, runner, hosts in bad_platforms:
            msg += f'\n * Platform {name} ({runner}) hosts: {", ".join(hosts)}'
        raise GlobalConfigError(msg)


def validate_platforms(platforms_cfg: Dict[str, Any]) -> None:
    """Check for invalid or inconsistent platforms config."""
    _validate_single_host(platforms_cfg)



#!/usr/bin/env python3
# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Cylc configuration linter.

Checks code style, deprecated syntax and other issues.

"""
# NOTE: docstring needed for `cylc help all` output and docs
# (if editing check this still comes out as expected)

COP_DOC = """cylc lint [OPTIONS] ARGS

Check .cylc and .rc files for code style, deprecated syntax and other issues.

By default, suggestions are written to stdout.

In-place mode ("-i, --inplace") writes suggestions into the file as comments.
Commit to version control before using this, in case you want to back out.

A non-zero return code will be returned if any issues are identified.
This can be overridden by providing the "--exit-zero" flag.
"""

NOQA = """
Individual errors can be ignored using the ``# noqa`` line comment.
It is good practice to specify specific errors you wish to ignore using
``# noqa: S002 S007 U999``
"""

TOMLDOC = """
pyproject.toml configuration:
   [tool.cylc.lint]
   ignore = ['S001', 'S002']    # List of rules to ignore
   exclude = ['etc/foo.cylc']   # List of files to ignore
   rulesets = ['style', '728']  # Sets default rulesets to check
   max-line-length = 130        # Max line length for linting
"""
import functools
from pathlib import Path
import re
import sys
import shutil
try:
    # BACK COMPAT: tomli
    #   Support for Python versions before tomllib was added to the
    #   standard library.
    # FROM: Python 3.7
    # TO: Python: 3.10
    from tomli import (
        loads as toml_loads,
        TOMLDecodeError,
    )
except ImportError:
    from tomllib import (  # type: ignore[no-redef]
        loads as toml_loads,
        TOMLDecodeError,
    )
from typing import (
    TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Union
)

from ansimarkup import parse as cparse

from cylc.flow import LOG
from cylc.flow.exceptions import CylcError
import cylc.flow.flags
from cylc.flow.loggingutil import set_timestamps
from cylc.flow.option_parsers import (
    CylcOptionParser as COP,
    WORKFLOW_ID_OR_PATH_ARG_DOC
)
from cylc.flow.cfgspec.workflow import upg, SPEC
from cylc.flow.id_cli import parse_id
from cylc.flow.parsec.config import ParsecConfig
from cylc.flow.scripts.cylc import DEAD_ENDS
from cylc.flow.terminal import cli_function

if TYPE_CHECKING:
    # BACK COMPAT: typing_extensions.Literal
    # FROM: Python 3.7
    # TO: Python 3.8
    from typing_extensions import Literal
    from optparse import Values

LINT_TABLE = ['tool', 'cylc', 'lint']
LINT_SECTION = '.'.join(LINT_TABLE)

# BACK COMPAT: DEPR_LINT_SECTION
# url:
#     https://github.com/cylc/cylc-flow/issues/5811
# from:
#    8.1.0
# to:
#    8.3.0
# remove at:
#    8.4.0 ?
DEPR_LINT_SECTION = 'cylc-lint'

IGNORE = 'ignore'
EXCLUDE = 'exclude'
RULESETS = 'rulesets'
MAX_LINE_LENGTH = 'max-line-length'

DEPRECATED_ENV_VARS = {
    'CYLC_SUITE_HOST': 'CYLC_WORKFLOW_HOST',
    'CYLC_SUITE_OWNER': 'CYLC_WORKFLOW_OWNER',
    'CYLC_SUITE_SHARE_DIR': 'CYLC_WORKFLOW_SHARE_DIR',
    'CYLC_SUITE_SHARE_PATH': 'CYLC_WORKFLOW_SHARE_PATH',
    'CYLC_SUITE_NAME': 'CYLC_WORKFLOW_ID',
    'CYLC_SUITE_LOG_DIR': 'CYLC_WORKFLOW_LOG_DIR',
    'CYLC_SUITE_INITIAL_CYCLE_POINT': 'CYLC_WORKFLOW_INITIAL_CYCLE_POINT',
    'CYLC_SUITE_INITIAL_CYCLE_TIME': 'CYLC_WORKFLOW_INITIAL_CYCLE_TIME',
    'CYLC_SUITE_FINAL_CYCLE_POINT': 'CYLC_WORKFLOW_FINAL_CYCLE_POINT',
    'CYLC_SUITE_FINAL_CYCLE_TIME': 'CYLC_WORKFLOW_FINAL_CYCLE_TIME',
    'CYLC_SUITE_WORK_DIR': 'CYLC_WORKFLOW_WORK_DIR',
    'CYLC_SUITE_UUID': 'CYLC_WORKFLOW_UUID',
    'CYLC_SUITE_RUN_DIR': 'CYLC_WORKFLOW_RUN_DIR',
}

OBSOLETE_ENV_VARS = {
    'CYLC_SUITE_DEF_PATH',
    'CYLC_SUITE_DEF_PATH_ON_SUITE_HOST'
}


DEPRECATED_STRING_TEMPLATES = {
    'suite': 'workflow',
    'suite_uuid': 'uuid',
    'batch_sys_name': 'job_runner_name',
    'batch_sys_job_id': 'job_id',
    'user@host': 'platform_name',
    'task_url': '``URL`` (if set in :cylc:conf:`[meta]URL`)',
    'workflow_url': (
        '``workflow_URL`` (if set in '
        ':cylc:conf:`[runtime][<namespace>][meta]URL`)'),
}


LIST_ITEM = '    * '


deprecated_string_templates = {
    key: (
        re.compile(r'%\(' + key + r'\)s'),
        value
    )
    for key, value in DEPRECATED_STRING_TEMPLATES.items()
}


def check_jinja2_no_shebang(
    line: str,
    file: Path,
    function: Callable,
    jinja_shebang: bool = False,
    **kwargs
):
    """Check ONLY top level workflow files for jinja without shebangs.

    Examples:
        >>> func = re.compile(r'{{').findall

        >>> check_jinja2_no_shebang(
        ... '{{FOO}}',
        ... function=func, jinja_shebang=True, file=Path('foo.cylc'))
        False

        >>> check_jinja2_no_shebang(
        ... '{{FOO}}',
        ... function=func, jinja_shebang=False, file=Path('suite.rc'))
        ['{{']
    """
    if (
        jinja_shebang
        or file.name not in {'flow.cylc', 'suite.rc'}
    ):
        return False
    return function(line)


def check_if_jinja2(
    line: str,
    jinja_shebang: bool,
    function: Callable,
    **kwargs
):
    """Run function if Jinja2 switched on.

    Examples:
        >>> func = re.compile('foo').findall

        >>> check_if_jinja2('barfoo', jinja_shebang=False, function=func)
        False

        >>> check_if_jinja2('foofoo', jinja_shebang=True, function=func)
        ['foo', 'foo']

    """
    if jinja_shebang:
        return function(line)
    return False


def check_dead_ends(line: str) -> bool:
    """Check for dead end cylc scripts as defined in cylc.flow.scripts.cylc

    Examples:
        # Context:
        # [runtime]
        #   [[task]]
        #     script = \"\"\"

        >>> check_dead_ends('        cylc check-software')
        True

        >>> check_dead_ends('        cylc log')
        False
    """
    return any(
        f'cylc {dead_end}' in line for dead_end in DEAD_ENDS
    )


def check_for_suicide_triggers(
    line: str,
    file: Path,
    function: Callable,
    **kwargs
):
    """Check for suicide triggers, if file is a .cylc file.

    Examples:
        >>> func = lambda line: line

        # Suicide trigger in a *.cylc file:
        >>> check_for_suicide_triggers(
        ... 'x:fail => !y', function=func, file=Path('foo.cylc'))
        'x:fail => !y'

        # Suicide trigger in a suite.rc file:
        >>> check_for_suicide_triggers(
        ... 'x:fail => !y', function=func, file=Path('suite.rc'))
        False
    """
    if file.name.endswith('.cylc'):
        return function(line)
    return False


def check_for_deprecated_environment_variables(
    line: str
) -> Union[bool, dict]:
    """Warn that environment variables with SUITE in are deprecated"""
    vars_found = [
        f'{k}: {v}' for k, v in DEPRECATED_ENV_VARS.items()
        if k in line
    ]

    if len(vars_found) == 1:
        return {'vars': vars_found}
    elif vars_found:
        return {'vars': '\n * ' + '\n * '.join(vars_found)}
    return False


def check_for_obsolete_environment_variables(line: str) -> List[str]:
    """Warn that environment variables are obsolete.

    Examples:

        >>> this = check_for_obsolete_environment_variables
        >>> this('CYLC_SUITE_DEF_PATH')
        ['CYLC_SUITE_DEF_PATH']
    """
    return [i for i in OBSOLETE_ENV_VARS if i in line]


def check_for_deprecated_task_event_template_vars(
    line: str
) -> Optional[Dict[str, str]]:
    """Look for string variables which are no longer supported

    Examples:
        >>> this = check_for_deprecated_task_event_template_vars

        >>> this('hello = "My name is %(suite)s"')
        {'list': '    * %(suite)s ⇒ %(workflow)s'}

        >>> expect = {'list': (
        ...     '    * %(suite)s ⇒ %(workflow)s    * %(task_url)s'
        ... ' - get ``URL`` (if set in :cylc:conf:`[meta]URL`)')}
        >>> this('hello = "My name is %(suite)s, %(task_url)s"') == expect
        True
    """
    result = []
    for key, (regex, replacement) in deprecated_string_templates.items():
        search_outcome = regex.findall(line)
        if search_outcome and ' ' in replacement:
            result.append(f'%({key})s - get {replacement}')
        elif search_outcome:
            result.append(f'%({key})s ⇒ %({replacement})s')

    if result:
        return {'list': LIST_ITEM + LIST_ITEM.join(result)}
    return None


INDENTATION = re.compile(r'^(\s*)(.*)')


def check_indentation(line: str) -> bool:
    """The key value pair is not indented 4*X spaces

    n.b. We test for trailing whitespace and incorrect section indenting
    elsewhere

    Examples:

        >>> check_indentation('')
        False
        >>> check_indentation('   ')
        False
        >>> check_indentation('   [')
        False
        >>> check_indentation('baz')
        False
        >>> check_indentation('    qux')
        False
        >>> check_indentation('   foo')
        True
        >>> check_indentation('     bar')
        True
    """
    match = INDENTATION.findall(line)[0]
    if not match[0] or not match[1] or match[1].startswith('['):
        return False
    return bool(len(match[0]) % 4 != 0)


INHERIT_REGEX = re.compile(r'\s*inherit\s*=\s*(.*)')
FAM_NAME_IGNORE_REGEX = re.compile(
    # Stuff we want to ignore when checking for lowercase in family names
    r'''
        # comments
        (?<!{)\#.*
        # or Cylc parameters
        | <[^>]+>
        # or Jinja2
        | {{.*?}} | {%.*?%} | {\#.*?\#}
        # or EmPy
        | (@[\[{\(]).*([\]\}\)])
    ''',
    re.X
)
LOWERCASE_REGEX = re.compile(r'[a-z]')


def check_lowercase_family_names(line: str) -> bool:
    """Check for lowercase in family names."""
    match = INHERIT_REGEX.match(line)
    if not match:
        return False
    # Replace stuff we want to ignore with a neutral char (tilde will do):
    content = FAM_NAME_IGNORE_REGEX.sub('~', match.group(1))
    return any(
        LOWERCASE_REGEX.search(i)
        for i in content.split(',')
        if i.strip(' \'"') not in {'None', 'none', 'root'}
    )


CHECK_FOR_OLD_VARS = re.compile(
    r'CYLC_VERSION\s*=\s*\{\{\s*CYLC_VERSION\s*\}\}'
    r'|ROSE_VERSION\s*=\s*\{\{\s*ROSE_VERSION\s*\}\}'
    r'|FCM_VERSION\s*=\s*\{\{\s*FCM_VERSION\s*\}\}'
)


def list_wrapper(line: str, check: Callable) -> Optional[Dict[str, str]]:
    """Take a line and a check function and return a Dict if there is a
    result, None otherwise.

    Returns

        Dict, in for {'vars': Error string}
    """
    result = check(line)
    if result:
        return {'vars': '\n    * '.join(result)}
    return None


FUNCTION = 'function'

STYLE_GUIDE = (
    'https://cylc.github.io/cylc-doc/stable/html/workflow-design-guide/'
    'style-guide.html#'
)
SECTION2 = r'\[\[\s*{}\s*\]\]'
SECTION3 = r'\[\[\[\s*{}\s*\]\]\]'
FILEGLOBS = ['*.rc', '*.cylc']
JINJA2_SHEBANG = '#!jinja2'
DEPENDENCY_SECTION_MSG = {
    'text': (
        '"[scheduling][dependencies][<recurrence>]graph =" -> '
        '"[scheduling][graph]<recurrence> ="'
    ),
    'rst': (
        '``[scheduling][dependencies][<recurrence>]graph =`` -> '
        '``[scheduling][graph]<recurrence> =``'
    )
}
JOBANDREMOTE_SECTION_MSG = {
    'text': (
        'settings in "[runtime][<namespace>][{}]" have been moved to '
        '"[runtime][<namespace>]" and "global.cylc[platforms]'
        '[<platforms name>]"'
    ),
    'rst': (
        'settings in ``[runtime][<namespace>][{}]`` have been moved to '
        '``[runtime][<namespace>]`` and ``global.cylc[platforms]'
        '[<platforms name>]``'
    )
}
JINJA2_FOUND_WITHOUT_SHEBANG = 'jinja2 found: no shebang (#!jinja2)'
CHECKS_DESC = {
    'U': '7 to 8 upgrades',
    'A': 'Auto Generated 7 to 8 upgrades',
    'S': 'Style'
}
LINE_LEN_NO = 'S012'
# Checks Dictionary fields:
# TODO: Consider making the checks an object.
# Key: A unique reference number.
# - short: A short description of the issue.
# - url: A link to a fuller description.
# - function: A function to use to run the check.
# - kwargs: We want to pass a set of common kwargs to the check function.
# - evaluate commented lines: Run this check on commented lines.
# - rst: An rst description, for use in the Cylc docs.
STYLE_CHECKS = {
    "S001": {
        'short': 'Use multiple spaces, not tabs',
        'url': STYLE_GUIDE + 'tab-characters',
        FUNCTION: re.compile(r'^\t').findall
    },
    "S002": {
        'short': 'Item not indented.',
        # Non-indented items should be sections:
        'url': STYLE_GUIDE + 'indentation',
        FUNCTION: re.compile(r'^[^%\{\[|\s]').findall
    },
    "S003": {
        'short': 'Top level sections should not be indented.',
        'url': STYLE_GUIDE + 'indentation',
        FUNCTION: re.compile(r'^\s+\[[^\[.]*\]').findall
    },
    "S004": {
        'short': (
            'Second level sections should be indented exactly '
            '4 spaces.'
        ),
        'url': STYLE_GUIDE + 'indentation',
        FUNCTION: re.compile(r'^(|\s|\s{2,3}|\s{5,})\[\[[^\[.]*\]\]').findall
    },
    "S005": {
        'short': (
            'Third level sections should be indented exactly '
            '8 spaces.'
        ),
        'url': STYLE_GUIDE + 'indentation',
        FUNCTION: re.compile(r'^(|\s{1,7}|\s{9,})\[\[\[[^\[.]*\]\]\]').findall
    },
    "S006": {
        'short': 'trailing whitespace.',
        'url': STYLE_GUIDE + 'trailing-whitespace',
        FUNCTION: re.compile(r'[ \t]$').findall
    },
    "S007": {
        'short': 'Family name contains lowercase characters.',
        'url': STYLE_GUIDE + 'task-naming-conventions',
        FUNCTION: check_lowercase_family_names,
    },
    "S008": {
        'short': JINJA2_FOUND_WITHOUT_SHEBANG,
        'url': '',
        'kwargs': True,
        FUNCTION: functools.partial(
            check_jinja2_no_shebang,
            function=re.compile(r'{[{%]').findall
        )
    },
    "S009": {
        'short': 'Host Selection Script may be redundant with platform',
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/7-to-8/'
            'major-changes/platforms.html'
        ),
        FUNCTION: re.compile(r'platform\s*=\s*\$\(.*?\)').findall,
    },
    "S010": {
        'short': 'Using backticks to invoke subshell is deprecated',
        'url': 'https://github.com/cylc/cylc-flow/issues/3825',
        FUNCTION: re.compile(r'platform\s*=\s*(`.*?`)').findall,
    },
    "S011": {
        'short': 'Cylc will process commented Jinja2!',
        'url': '',
        'kwargs': True,
        'evaluate commented lines': True,
        FUNCTION: functools.partial(
            check_if_jinja2,
            function=re.compile(r'(?<!{)#[^$].*?{[{%]').findall
        )
    },
    'S012': {
        'short': 'This number is reserved for line length checks',
    },
    'S013': {
        'short': 'Items should be indented in 4 space blocks.',
        FUNCTION: check_indentation
    }
}
# Subset of deprecations which are tricky (impossible?) to scrape from the
# upgrader.
MANUAL_DEPRECATIONS = {
    "U001": {
        'short': (
            DEPENDENCY_SECTION_MSG['text'] + ' (``[dependencies]`` detected)'
        ),
        'url': '',
        'rst': (
            DEPENDENCY_SECTION_MSG['rst'] + ' (``[dependencies]`` detected)'
        ),
        FUNCTION: re.compile(SECTION2.format('dependencies')).findall,
    },
    "U002": {
        'short': DEPENDENCY_SECTION_MSG['text'] + ' (``graph =`` detected)',
        'url': '',
        'rst': DEPENDENCY_SECTION_MSG['rst'] + ' (``graph =`` detected)',
        FUNCTION: re.compile(r'graph\s*=\s*').findall,
    },
    "U003": {
        'short': JOBANDREMOTE_SECTION_MSG['text'].format('remote'),
        'url': '',
        'rst': JOBANDREMOTE_SECTION_MSG['rst'].format('remote'),
        FUNCTION: re.compile(SECTION3.format('remote')).findall
    },
    "U004": {
        'short': JOBANDREMOTE_SECTION_MSG['text'].format('job'),
        'url': '',
        'rst': JOBANDREMOTE_SECTION_MSG['rst'].format('job'),
        FUNCTION: re.compile(SECTION3.format('job')).findall,
    },
    "U005": {
        'short': (
            'flow.cylc[runtime][<namespace>][job]batch system -> '
            'global.cylc[platforms][<platform name>]job runner'
        ),
        'url': '',
        'rst': (
            '``flow.cylc[runtime][<namespace>][job]batch system`` -> '
            '``global.cylc[platforms][<platform name>]job runner``'
        ),
        FUNCTION: re.compile(r'batch system\s*=\s*').findall,
    },
    "U006": {
        'short': 'Using backticks to invoke subshell will fail at Cylc 8.',
        'url': 'https://github.com/cylc/cylc-flow/issues/3825',
        FUNCTION: re.compile(r'host\s*=\s*(`.*?`)').findall,
    },
    'U007': {
        'short': (
            'Use built in platform selection instead of rose host-select.'),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/7-to-8/'
            'major-changes/platforms.html'),
        FUNCTION: re.compile(r'platform\s*=\s*\$\(\s*rose host-select').findall
    },
    'U008': {
        'short': 'Suicide triggers are not required at Cylc 8.',
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/7-to-8'
            '/major-changes/suicide-triggers.html'),
        'kwargs': True,
        FUNCTION: functools.partial(
            check_for_suicide_triggers,
            function=re.compile(r'=>\s*\!.*').findall
        ),
    },
    'U009': {
        'short': 'This line contains an obsolete Cylc CLI command.',
        'url': '',
        FUNCTION: check_dead_ends
    },
    'U010': {
        'short': 'rose suite-hook is deprecated at Rose 2,',
        'url': '',
        FUNCTION: lambda line: 'rose suite-hook' in line,
    },
    'U011': {
        'short': 'Leading zeros are no longer valid for Jinja2 integers.',
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/7-to-8/major-changes'
            '/python-2-3.html#jinja2-integers-with-leading-zeros'),
        'kwargs': True,
        FUNCTION: functools.partial(
            check_if_jinja2,
            function=re.compile(r'\{%\s*set\s*.+?\s*=\s*0\d+\s*%\}').findall
        )
    },
    'U012': {
        'short': (
            'Deprecated environment variables: {vars}'),
        'rst': (
            'The following environment variables are deprecated:\n\n'
            '.. list-table::'
            '\n   :header-rows: 1'
            '\n\n   * - Deprecated Variable'
            '\n     - New Variable'
        ) + ''.join(
            [
                f'\n   * - ``{old}``\n     - ``{new}``'
                for old, new in DEPRECATED_ENV_VARS.items()
            ]
        ),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/reference/'
            'job-script-vars/index.html'
        ),
        FUNCTION: check_for_deprecated_environment_variables,
    },
    'U013': {
        'short': (
            'Obsolete environment variables: {vars}'),
        'rst': (
            'The following environment variables are obsolete:\n\n'
            + ''.join([f'\n * ``{old}``' for old in OBSOLETE_ENV_VARS])
        ),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/reference/'
            'job-script-vars/index.html'
        ),
        FUNCTION: check_for_obsolete_environment_variables,
    },
    'U014': {
        'short': 'Use "isodatetime [ref]" instead of "rose date [-c]"',
        'rst': (
            'For datetime operations in task scripts:\n\n'
            ' * Use ``isodatetime`` instead of ``rose date``\n'
            ' * Use ``isodatetime ref`` instead of ``rose date -c`` for '
            'the current cycle point\n'
        ),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/7-to-8/'
            'cheat-sheet.html#datetime-operations'
        ),
        FUNCTION: re.compile(r'rose +date').findall,
    },
    'U015': {
        'short': (
            'Deprecated template variables.'),
        'rst': (
            'The following template variables, mostly used in event handlers,'
            'are deprecated, and should be replaced:'
            + ''.join([
                f'\n * ``{old}`` ⇒ {new}'
                for old, new in DEPRECATED_STRING_TEMPLATES.items()
            ])
        ),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/user-guide/'
            'writing-workflows/runtime.html#task-event-template-variables'
        ),
        FUNCTION: check_for_deprecated_task_event_template_vars,
    },
    'U016': {
        'short': 'Deprecated template vars: {vars}',
        'rst': (
            'It is no longer necessary to configure the environment variables '
            '``CYLC_VERSION``, ``ROSE_VERSION`` or ``FCM_VERSION``.'
        ),
        'url': (
            'https://cylc.github.io/cylc-doc/stable/html/plugins/'
            'cylc-rose.html#special-variables'
        ),
        FUNCTION: functools.partial(
            list_wrapper, check=CHECK_FOR_OLD_VARS.findall),
    },
}
ALL_RULESETS = ['728', 'style', 'all']
EXTRA_TOML_VALIDATION = {
    IGNORE: {
        lambda x: re.match(r'[A-Z]\d\d\d', x):
            '{item} not valid: Ignore codes should be in the form X001',
        lambda x: x in parse_checks(['728', 'style']):
            '{item} is a not a known linter code.'
    },
    RULESETS: {
        lambda item: item in ALL_RULESETS:
            '{item} not valid: Rulesets can be '
            '\'728\', \'style\' or \'all\'.'
    },
    MAX_LINE_LENGTH: {
        lambda x: isinstance(x, int):
            'max-line-length must be an integer.'
    },
    # consider checking that item is file?
    EXCLUDE: {}
}


def parse_ruleset_option(ruleset: str) -> List[str]:
    if ruleset in {'all', ''}:
        return ['728', 'style']
    return [ruleset]


def get_url(check_meta: Dict) -> str:
    """Get URL from check data.

    If the URL doesn't start with http then prepend with address
    of the 7-to-8 upgrade guide.

    Examples:
        >>> get_url({'no': 'url key'})
        ''
        >>> get_url({'url': ''})
        ''
        >>> get_url({'url': 'https://www.h2g2.com/'})
        'https://www.h2g2.com/'
        >>> get_url({'url': 'cheat-sheet.html'})
        'https://cylc.github.io/cylc-doc/stable/html/7-to-8/cheat-sheet.html'
    """
    url = check_meta.get('url', '')
    if url and not url.startswith('http'):
        url = (
            "https://cylc.github.io/cylc-doc/stable/html/7-to-8/"
            + check_meta['url']
        )
    return url


def validate_toml_items(tomldata):
    """Check that all tomldata items are lists of strings

    Explicitly checks and raises if tomldata not:
        - str in EXTRA_TOML_VALIDATION.keys()
        - item is not a list of strings.

    Plus additional validation for each set of values.
    """
    for key, items in tomldata.items():
        # Key should only be one of the allowed keys:
        if key not in EXTRA_TOML_VALIDATION.keys():
            raise CylcError(
                f'Only {[*EXTRA_TOML_VALIDATION.keys()]} '
                f'allowed as toml sections but you used "{key}"'
            )
        if key != MAX_LINE_LENGTH:
            # Item should be a list...
            if not isinstance(items, list):
                raise CylcError(
                    f'{key} should be a list, but was: {items}')
            # ... of strings
            for item in items:
                if not isinstance(item, str):
                    raise CylcError(
                        f'Config {item} should be a string but '
                        f'is {str(type(item))}'
                    )
                for check, message in EXTRA_TOML_VALIDATION[key].items():
                    if not check(item):
                        raise CylcError(
                            message.format(item=item)
                        )
    return True


def get_pyproject_toml(dir_: Path) -> Dict[str, Any]:
    """if a pyproject.toml file is present open it and return settings.
    """
    tomlfile = dir_ / 'pyproject.toml'
    tomldata: Dict[str, Union[List[str], int, None]] = {
        RULESETS: [],
        IGNORE: [],
        EXCLUDE: [],
        MAX_LINE_LENGTH: None,
    }
    if tomlfile.is_file():
        try:
            loadeddata = toml_loads(tomlfile.read_text())
        except TOMLDecodeError as exc:
            raise CylcError(f'pyproject.toml did not load: {exc}')

        _tool, _cylc, _lint = LINT_TABLE
        try:
            data = loadeddata[_tool][_cylc][_lint]
        except KeyError:
            if DEPR_LINT_SECTION in loadeddata:
                LOG.warning(
                    f"The [{DEPR_LINT_SECTION}] section in pyproject.toml is "
                    f"deprecated. Use [{LINT_SECTION}] instead."
                )
            data = loadeddata.get(DEPR_LINT_SECTION, {})
        tomldata.update(data)
        validate_toml_items(tomldata)

    return tomldata


def merge_cli_with_tomldata(target: Path, options: 'Values') -> Dict[str, Any]:
    """Get a list of checks based on the checking options

    Args:
        target: Location being linted, in which we might find a
            pyproject.toml file.
        options: Cli Options

    This has not been merged with merged with the logic in
    _merge_cli_with_tomldata to keep the testing of file-system touching
    and pure logic separate.
    """
    ruleset_default = (options.ruleset == '')
    options.ruleset = parse_ruleset_option(options.ruleset)
    tomlopts = get_pyproject_toml(target)
    return _merge_cli_with_tomldata(
        {
            EXCLUDE: [],
            IGNORE: options.ignores,
            RULESETS: options.ruleset
        },
        tomlopts,
        ruleset_default
    )


def _merge_cli_with_tomldata(
    clidata: Dict[str, Any],
    tomldata: Dict[str, Any],
    override_cli_default_rules: bool = False
) -> Dict[str, Any]:
    """Merge options set by pyproject.toml with CLI options.

    rulesets: CLI should override toml.
    ignore: CLI and toml should be combined with no duplicates.
    exclude: No CLI equivalent, return toml if any.

    Args:
        override_cli_default_rules: If user doesn't specifiy a ruleset use the
            rules from the tomlfile - i.e: if we've set 'rulesets': 'style'
            we probably don't want to get warnings about 728 upgrades by
            default, but only if we ask for it on the CLI.

    Examples:
    >>> result = _merge_cli_with_tomldata(
    ... {'rulesets': ['foo'], 'ignore': ['R101'], 'exclude': []},
    ... {'rulesets': ['bar'], 'ignore': ['R100'], 'exclude': ['*.bk']})
    >>> result['ignore']
    ['R100', 'R101']
    >>> result['rulesets']
    ['foo']
    >>> result['exclude']
    ['*.bk']
    """
    if isinstance(clidata[RULESETS][0], list):
        clidata[RULESETS] = clidata[RULESETS][0]

    output = {}

    # Combine 'ignore' sections:
    output[IGNORE] = sorted(set(clidata[IGNORE] + tomldata[IGNORE]))

    # Replace 'rulesets' from toml with those from CLI if they exist:

    if override_cli_default_rules:
        output[RULESETS] = (
            tomldata[RULESETS] if tomldata[RULESETS]
            else clidata[RULESETS]
        )
    else:
        output[RULESETS] = (
            clidata[RULESETS] if clidata[RULESETS]
            else tomldata[RULESETS]
        )

    # Return 'exclude' and 'max-line-length' for the tomldata:
    output[EXCLUDE] = tomldata[EXCLUDE]
    output[MAX_LINE_LENGTH] = tomldata.get(MAX_LINE_LENGTH, None)

    return output


def list_to_config(path_, is_section=False):
    """Prettify a config list

    Args:
        path_: list forming address in the config.
        is_section: Is this item a section?

    Examples:
        >>> list_to_config(['foo', 'bar'], False)
        '[foo]bar'

        >>> list_to_config(['foo', 'bar'], True)
        '[foo][bar]'
    """
    output = ''
    for item in path_[:-1]:
        # All but the last item must be sections:
        output += f'[{item}]'
    if is_section:
        # Last item is section:
        output += f'[{path_[-1]}]'
    else:
        # Last item not a section:
        output += path_[-1]
    return output


def get_upgrader_info():
    """Extract info about obseletions and deprecations from Parsec Objects."""
    conf = ParsecConfig(SPEC, upg)
    upgrades = conf.upgrader(conf.dense, '').upgrades
    deprecations = {}

    for upgrades_for_version in upgrades.values():
        for index, upgrade in enumerate(upgrades_for_version):
            # Set a flag indicating that a variable has been moved.
            is_dep, is_obs = False, False
            if upgrade['new'] is None:
                section_name = list_to_config(
                    upgrade["old"], upgrade["is_section"])
                short = f'{section_name} - not available at Cylc 8'
                rst = f'``{section_name}`` is not available at Cylc 8'
                is_obs = True
            elif upgrade["old"][-1] == upgrade['new'][-1]:
                # Where an item with the same name has been moved
                # a 1 line regex isn't going to work.
                continue
            else:
                old = list_to_config(
                    upgrade["old"], upgrade["is_section"])
                new = list_to_config(
                    upgrade["new"], upgrade["is_section"])
                short = f'{old} -> {new}'
                rst = f'``{old}`` is now ``{new}``'
                is_dep = True

            # Check whether upgrade is section:
            if upgrade['is_section'] is True:
                section_depth = len(upgrade['old'])
                start = r'\[' * section_depth
                end = r'\]' * section_depth
                name = upgrade["old"][-1]
                expr = fr'{start}\s*{name}\s*{end}\s*$'
            else:
                name = upgrade["old"][-1]
                expr = rf'^\s*{name}\s*=\s*.*'

            deprecations[f'A{index:03d}'] = {
                'short': short,
                'url': '',
                'rst': rst,
                FUNCTION: re.compile(expr).findall,
                'is_obs': is_obs,
                'is_dep': is_dep,
            }

    return deprecations


PURPOSE_FILTER_MAP = {
    'style': 'S',
    '728': 'UA',
}


def parse_checks(check_args, ignores=None, max_line_len=None, reference=False):
    """Prepare dictionary of checks.

    Args:
        check_arg: list of types of checks to run,
            currently expecting '728' and/or 'style'
        ignores: list of codes to ignore.
        max_line_len: Adds a specific style warning for lines longer than
            this. (If None, rule not enforced)
        reference: Function is being used to get a reference. If true
            max-line-length will have a generic message, rather than
            using any specific value.
    """
    ignores = ignores or []
    parsedchecks = {}
    purpose_filters = [
        purpose
        for arg in check_args
        for purpose in PURPOSE_FILTER_MAP[arg]
    ]

    checks = {
        'S': STYLE_CHECKS,
        'U': MANUAL_DEPRECATIONS,
        'A': get_upgrader_info(),
    }
    for purpose, ruleset in checks.items():
        if purpose in purpose_filters:
            # Run through the rest of the config items.
            for index, meta in ruleset.items():
                meta.update({'purpose': purpose})
                if f'{index}' not in ignores:
                    parsedchecks.update({index: meta})
            if 'S' in purpose and LINE_LEN_NO not in ignores:
                # Special handling for max line length:
                if not max_line_len:
                    max_line_len = 130
                regex = r"^.{" + str(max_line_len) + r"}"
                if reference:
                    msg = (
                        'line > ``<max_line_len>`` characters. Max line '
                        ' length set in pyproject.toml (default 130)'
                    )
                else:
                    msg = f'line > {max_line_len} characters.'
                parsedchecks[LINE_LEN_NO] = {
                    'short': msg,
                    'url': STYLE_GUIDE + 'line-length-and-continuation',
                    FUNCTION: re.compile(regex).findall,
                    'purpose': 'S'
                }
    return parsedchecks


def get_index_str(meta: dict, index: str) -> str:
    """Printable purpose string - mask useless numbers for auto-generated
    upgrades."""
    if meta.get('is_dep', None):
        return 'U998'
    elif meta.get('is_obs', None):
        return 'U999'
    else:
        return f'{index}'


def check_cylc_file(
    file: Path,
    file_rel: Path,
    checks: Dict[str, dict],
    counter: Dict[str, int],
    modify: bool = False,
):
    """Check A Cylc File for Cylc 7 Config"""
    with open(file, 'r') as cylc_file:
        # generator which reads and lints one line at a time
        linter = lint(
            file_rel,
            cylc_file,
            checks,
            counter,
            modify,
        )

        if modify:
            # write modifications into a ".temp" file
            modify_file_path = file.parent / f'{file.name}.temp'
            with open(modify_file_path, 'w+') as modify_file:
                for line in linter:
                    modify_file.write(line)
            # replace the original with the ".temp" file
            shutil.move(str(modify_file_path), file)
        else:
            for _line in linter:
                pass


def no_qa(line: str, index: str):
    """This line has a no-qa comment.

    Examples:
        # No comment, no exception:
        >>> no_qa('foo = bar', 'S001')
        False

        # Comment, no error codes, no checking:
        >>> no_qa('foo = bar # noqa', 'S001')
        True

        # Comment, no relevent error codes, no checking:
        >>> no_qa('foo = bar # noqa: S999, 997', 'S001')
        False

        # Comment, relevent error codes, checking:
        >>> no_qa('foo = bar # noqa: S001 S003', 'S001')
        True
    """
    NOQA = re.compile(r'.*#\s*[Nn][Oo][Qq][Aa]:?(.*)')
    noqa = NOQA.findall(line)
    if noqa and (noqa[0] == '' or index in noqa[0]):
        return True
    return False


def lint(
    file_rel: Path,
    lines: Iterator[str],
    checks: Dict[str, dict],
    counter: Dict[str, int],
    modify: bool = False,
    write: Callable = print
) -> Iterator[str]:
    """Lint text, one line at a time.

    Arguments:
        file_rel:
            The filepath relative to the workflow configuration directory
            (used in messages).
        lines:
            Iterator which produces one line of text at a time
            e.g. open(file) or iter(['foo\n', 'bar\n', 'baz\n'].
        counter:
            Dictionary for counting lint hits per category.
        modify:
            If True, this generator will yield the file one line at a time
            with comments inserted to help users fix their lint.
        write:
            A function for reporting lint messages.

    Yields:
        The original file with added comments when `modify is True`.

    """
    # get the first line
    line_no = 1
    line = next(lines)
    # check if it is a jinja2 shebang
    jinja_shebang = line.strip().lower() == JINJA2_SHEBANG

    while True:
        # run lint checks against the current line
        for index, check_meta in checks.items():
            # Skip commented line unless check says not to.
            index_str = get_index_str(check_meta, index)
            if (
                (
                    line.strip().startswith('#')
                    and not check_meta.get('evaluate commented lines', False)
                )
                or no_qa(line, index_str)
            ):
                continue

            if check_meta.get('kwargs', False):
                # Use a more complex function with keywords:
                check_function = functools.partial(
                    check_meta['function'],
                    check_meta=check_meta,
                    file=file_rel,
                    jinja_shebang=jinja_shebang,
                )
            else:
                # Just going to pass the line to the check function:
                check_function = check_meta['function']

            # Run the check:
            check = check_function(line)

            if check:
                # we have lint!
                if isinstance(check, dict):
                    msg = check_meta['short'].format(**check)
                else:
                    msg = check_meta['short']
                counter.setdefault(check_meta['purpose'], 0)
                counter[check_meta['purpose']] += 1
                if modify:
                    # insert a command to help the user
                    url = get_url(check_meta)

                    yield (
                        f'# [{index_str}]: '
                        f'{msg}\n'
                        f'# - see {url}\n'
                    )
                else:
                    # write a message to inform the user
                    write(cparse(
                        '<yellow>'
                        f'[{index_str}] {file_rel}:{line_no}: {msg}'
                        '</yellow>'
                    ))
        if modify:
            yield line

        try:
            # get the next line
            line = next(lines)
        except StopIteration:
            # end of interator
            return
        line_no += 1


def get_cylc_files(
    base: Path, exclusions: Union[list, None] = None
) -> Iterator[Path]:
    """Given a directory yield paths to check."""
    exclusions = [] if exclusions is None else exclusions
    except_these_files = [
        file for exclusion in exclusions for file in base.rglob(exclusion)]
    for rglob in FILEGLOBS:
        for path in base.rglob(rglob):
            # Exclude log directory:
            if (
                path.relative_to(base).parts[0] != 'log'
                and path not in except_these_files
            ):
                yield path


def get_reference(ruleset: str, output_type: 'Literal["text", "rst"]') -> str:
    """Fill out a template with all the issues Cylc Lint looks for.
    """
    checks = parse_checks(
        parse_ruleset_option(ruleset),
        reference=True
    )

    issue_heading_template = (
        '\n{url}_\n{underline}\n{summary}\n\n' if output_type == 'rst' else
        '\n{check}:\n    {summary}\n    {url}\n\n'
    )
    output = ''
    current_checkset = ''
    for index, meta in checks.items():
        # Check if the purpose has changed - if so create a new
        # section heading:
        if meta['purpose'] != current_checkset:
            current_checkset = meta['purpose']
            title = CHECKS_DESC[meta["purpose"]]
            output += '\n{title}\n{underline}\n'.format(
                title=title, underline="-" * len(title)
            )

            if current_checkset == 'A':
                output += (
                    'U998 and U999 represent automatically generated'
                    ' sets of deprecations and upgrades.'
                )

        # Fill a template with info about the issue.
        if output_type == 'rst':
            summary = meta.get("rst", meta['short'])
        elif output_type == 'text':
            summary = meta.get("short").replace('``', '')

        if current_checkset == 'A':
            # Condensed check summary for auto-generated lint items.
            if output_type == 'rst':
                output += '\n'
            output += '\n* ' + summary
        else:
            check = get_index_str(meta, index)
            template = issue_heading_template
            url = get_url(meta)
            if output_type == 'rst':
                url = f'`{check} <{url}>`' if url else f'{check}'
            msg = template.format(
                title=index,
                check=check,
                summary=summary,
                url=url,
                underline=(len(url) + 1) * '^'
            )
            output += msg
    output += '\n'
    return output


def target_version_check(
    target: Path,
    quiet: 'Values',
    mergedopts: Dict[str, Any]
) -> List:
    """
    Check whether target is an upgraded Cylc 8 workflow.

    If it isn't then we shouldn't run the 7-to-8 checks upon
    it.

    If it isn't and the only ruleset requested by the user is '728'
    we should exit with an error code unless the user has specifically
    disabled thatr with --exit-zero.
    """
    cylc8 = (target / 'flow.cylc').exists()
    if not cylc8 and mergedopts[RULESETS] == ['728']:
        LOG.error(
            f'{target} not a Cylc 8 workflow: '
            'Lint after renaming '
            '"suite.rc" to "flow.cylc"'
        )
        sys.exit(not quiet)
    elif not cylc8 and '728' in mergedopts[RULESETS]:
        check_names = mergedopts[RULESETS]
        check_names.remove('728')
    else:
        check_names = mergedopts[RULESETS]
    return check_names


def get_option_parser() -> COP:
    parser = COP(
        (
            COP_DOC
            + NOQA.replace('``', '"')
            + TOMLDOC
        ),
        argdoc=[
            COP.optional(WORKFLOW_ID_OR_PATH_ARG_DOC)
        ],
    )
    parser.add_option(
        '--inplace', '-i',
        help=(
            'Modify files in place, adding comments to files. '
            'If not set, the script will work as a linter'
        ),
        action='store_true',
        default=False,
    )
    parser.add_option(
        '--ruleset', '-r',
        help=(
            'Set of rules to use: '
            '("728", "style", "all")'
        ),
        default='',
        choices=["728", "style", "all", ''],
        dest='ruleset'
    )
    parser.add_option(
        '--list-codes',
        help=(
            'List all linter codes.'
        ),
        action='store_true',
        default=False,
        dest='ref_mode'
    )
    parser.add_option(
        '--ignore', '-n',
        help=(
            'Ignore this check number.'
        ),
        action='append',
        default=[],
        dest='ignores',
        metavar="CODE",
        choices=list(STYLE_CHECKS.keys()) + [LINE_LEN_NO]
    )
    parser.add_option(
        '--exit-zero',
        help='Exit with status code "0" even if there are issues.',
        action='store_true',
        default=False,
        dest='exit_zero'
    )

    return parser


@cli_function(get_option_parser)
def main(parser: COP, options: 'Values', target=None) -> None:
    if cylc.flow.flags.verbosity < 2:
        set_timestamps(LOG, False)

    if options.ref_mode:
        print(get_reference(options.ruleset, 'text'))
        sys.exit(0)

    # If target not given assume we are looking at PWD:
    if target is None:
        target = str(Path.cwd())

    # make sure the target is a src/run directory:
    _, _, target = parse_id(
        target,
        src=True,
        constraint='workflows',
    )

    # We want target to be the containing folder, not the flow.cylc
    # file identified by parse_id:
    target = target.parent

    mergedopts = merge_cli_with_tomldata(target, options)

    check_names = target_version_check(
        target=target, quiet=options.exit_zero, mergedopts=mergedopts)

    # Get the checks object.
    checks = parse_checks(
        check_names,
        ignores=mergedopts[IGNORE],
        max_line_len=mergedopts[MAX_LINE_LENGTH]
    )

    # Check each file matching a pattern:
    counter: Dict[str, int] = {}
    for file in get_cylc_files(target, mergedopts[EXCLUDE]):
        LOG.debug(f'Checking {file}')
        check_cylc_file(
            file,
            file.relative_to(target),
            checks,
            counter,
            options.inplace,
        )

    if counter:
        total_lint_hits = sum(counter.values())
        msg = cparse(
            '\n<yellow>'
            f'Checked {target} against {check_names} '
            f'rules and found {total_lint_hits} issue'
            f'{"s" if total_lint_hits > 1 else ""}.'
            '</yellow>'
        )
    else:
        msg = cparse(
            '<green>'
            f'Checked {target} against {check_names} rules and '
            'found no issues.'
            '</green>'
        )

    print(msg)

    # Exit with an error code if there were warnings and
    # if --exit-zero was not set.
    # Return codes: sys.exit(True) == 1, sys.exit(False) == 0
    sys.exit(bool(counter) and not options.exit_zero)


# NOTE: use += so that this works with __import__
# (docstring needed for `cylc help all` output)
__doc__ += NOQA
__doc__ += get_reference('all', 'rst')



#!/usr/bin/env python3
# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Common utilities for Tui."""

from contextlib import contextmanager
from getpass import getuser
from itertools import zip_longest
import re
from time import time
from typing import Tuple

from cylc.flow import LOG
from cylc.flow.id import Tokens
from cylc.flow.task_state import (
    TASK_STATUS_RUNNING
)
from cylc.flow.tui import (
    JOB_COLOURS,
    JOB_ICON,
    TASK_ICONS,
    TASK_MODIFIERS
)
from cylc.flow.wallclock import get_unix_time_from_time_string


# the Tui user, note this is always the same as the workflow owner
# (Tui doesn't do multi-user stuff)
ME = getuser()


@contextmanager
def suppress_logging():
    """Suppress Cylc logging.

    Log goes to stdout/err which can pollute Urwid apps.
    Patching sys.stdout/err is insufficient so we set the level to something
    silly for the duration of this context manager then set it back again
    afterwards.
    """
    level = LOG.getEffectiveLevel()
    LOG.setLevel(99999)
    yield
    LOG.setLevel(level)


def get_task_icon(
    status,
    *,
    is_held=False,
    is_queued=False,
    is_runahead=False,
    start_time=None,
    mean_time=None
):
    """Return a Unicode string to represent a task.

    Arguments:
        status (str):
            A Cylc task status string.
        is_held (bool):
            True if the task is held.
        is_queued (bool):
            True if the task is queued.
        is_runahead (bool):
            True if the task is runahead limited.
        start_time (str):
            Start date time string.
        mean_time (int):
            Execution mean time.

    Returns:
        list - Text content for the urwid.Text widget,
        may be a string, tuple or list, see urwid docs.

    """
    ret = []
    if is_held:
        ret.append(TASK_MODIFIERS['held'])
    elif is_runahead:
        ret.append(TASK_MODIFIERS['runahead'])
    elif is_queued:
        ret.append(TASK_MODIFIERS['queued'])
    if (
        status == TASK_STATUS_RUNNING
        and start_time
        and mean_time
    ):
        start_time = get_unix_time_from_time_string(start_time)
        progress = (time() - start_time) / mean_time
        if progress >= 0.75:
            status = f'{TASK_STATUS_RUNNING}:75'
        elif progress >= 0.5:
            status = f'{TASK_STATUS_RUNNING}:50'
        elif progress >= 0.25:
            status = f'{TASK_STATUS_RUNNING}:25'
        else:
            status = f'{TASK_STATUS_RUNNING}:0'
    ret.append(TASK_ICONS[status])
    return ret


def idpop(id_):
    """Remove the last element of a node id.

    Examples:
        >>> idpop('w//c/t/j')
        'w//c/t'
        >>> idpop('c/t/j')
        '//c/t'
        >>> idpop('c/t')
        '//c'
        >>> idpop('c')
        Traceback (most recent call last):
        ValueError: No tokens provided
        >>> idpop('')
        Traceback (most recent call last):
        ValueError: Invalid Cylc identifier: //

    """
    relative = '//' not in id_
    tokens = Tokens(id_, relative=relative)
    tokens.pop_token()
    return tokens.id


def compute_tree(data):
    """Digest GraphQL data to produce a tree."""
    root_node = add_node('root', 'root', {}, data={})

    for flow in data['workflows']:
        nodes = {}
        flow_node = add_node(
            'workflow', flow['id'], nodes, data=flow)
        root_node['children'].append(flow_node)

        # populate cycle nodes
        for cycle in flow.get('cyclePoints', []):
            cycle['id'] = idpop(cycle['id'])  # strip the family off of the id
            cycle_node = add_node('cycle', cycle['id'], nodes, data=cycle)
            flow_node['children'].append(cycle_node)

        # populate family nodes
        for family in flow.get('familyProxies', []):
            add_node('family', family['id'], nodes, data=family)

        # create cycle/family tree
        for family in flow.get('familyProxies', []):
            family_node = add_node(
                'family', family['id'], nodes)
            first_parent = family['firstParent']
            if (
                first_parent
                and first_parent['name'] != 'root'
            ):
                parent_node = add_node(
                    'family', first_parent['id'], nodes)
                parent_node['children'].append(family_node)
            else:
                add_node(
                    'cycle', idpop(family['id']), nodes
                )['children'].append(family_node)

        # add leaves
        for task in flow.get('taskProxies', []):
            # If there's no first parent, the child will have been deleted
            # during/after API query resolution. So ignore.
            if not task['firstParent']:
                continue
            task_node = add_node(
                'task', task['id'], nodes, data=task)
            if task['firstParent']['name'] == 'root':
                family_node = add_node(
                    'cycle', idpop(task['id']), nodes)
            else:
                family_node = add_node(
                    'family', task['firstParent']['id'], nodes)
            family_node['children'].append(task_node)
            for job in task['jobs']:
                job_node = add_node(
                    'job', job['id'], nodes, data=job)
                job_info_node = add_node(
                    'job_info', job['id'] + '_info', nodes, data=job)
                job_node['children'] = [job_info_node]
                task_node['children'].append(job_node)

        # sort
        for (type_, _), node in nodes.items():
            if type_ != 'task':
                # NOTE: jobs are sorted by submit-num in the GraphQL query
                node['children'].sort(
                    key=lambda x: NaturalSort(x['id_'])
                )

        # spring nodes
        if 'port' not in flow:
            # the "port" field is only available via GraphQL
            # so we are not connected to this workflow yet
            flow_node['children'].append(
                add_node(
                    '#spring',
                    '#spring',
                    nodes,
                    data={
                        'id': flow.get('_tui_data', 'Loading ...'),
                    }
                )
            )

    return root_node


class NaturalSort:
    """An object to use as a sort key for sorting strings as a human would.

    This recognises numerical patterns within strings.

    Examples:
        >>> N = NaturalSort

        String comparisons work as normal:
        >>> N('') < N('')
        False
        >>> N('a') < N('b')
        True
        >>> N('b') < N('a')
        False

        Integer comparisons work as normal:
        >>> N('9') < N('10')
        True
        >>> N('10') < N('9')
        False

        Integers rank higher than strings:
        >>> N('1') < N('a')
        True
        >>> N('a') < N('1')
        False

        Integers within strings are sorted numerically:
        >>> N('a9b') < N('a10b')
        True
        >>> N('a10b') < N('a9b')
        False

        Lexicographical rules apply when substrings match:
        >>> N('a1b2') < N('a1b2c3')
        True
        >>> N('a1b2c3') < N('a1b2')
        False

        Equality works as per regular string rules:
        >>> N('a1b2c3') == N('a1b2c3')
        True

    """

    PATTERN = re.compile(r'(\d+)')

    def __init__(self, value):
        self.value = tuple(
            int(item) if item.isdigit() else item
            for item in self.PATTERN.split(value)
            # remove empty strings if value ends with a digit
            if item
        )

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        for this, that in zip_longest(self.value, other.value):
            if this is None:
                return True
            if that is None:
                return False
            this_isstr = isinstance(this, str)
            that_isstr = isinstance(that, str)
            if this_isstr and that_isstr:
                if this == that:
                    continue
                return this < that
            this_isint = isinstance(this, int)
            that_isint = isinstance(that, int)
            if this_isint and that_isint:
                if this == that:
                    continue
                return this < that
            if this_isint and that_isstr:
                return True
            if this_isstr and that_isint:
                return False
        return False


def dummy_flow(data):
    return add_node(
        'workflow',
        data['id'],
        {},
        data
    )


def add_node(type_, id_, nodes, data=None):
    """Create a node add it to the store and return it.

    Arguments:
        type_ (str):
            A string to represent the node type.
        id_ (str):
            A unique identifier for this node.
        nodes (dict):
            The node store to add the new node to.
        data (dict):
            An optional dictionary of data to add to this node.
            Can be left to None if retrieving a node from the store.

    Returns:
        dict - The requested node.

    """
    if (type_, id_) not in nodes:
        nodes[(type_, id_)] = {
            'children': [],
            'id_': id_,
            'data': data or {},
            'type_': type_
        }
    return nodes[(type_, id_)]


def get_job_icon(status):
    """Return a unicode string to represent a job.

    Arguments:
        status (str): A Cylc job status string.

    Returns:
        list - Text content for the urwid.Text widget,
        may be a string, tuple or list, see urwid docs.

    """
    return [
        (f'job_{status}', JOB_ICON)
    ]


def get_task_status_summary(flow):
    """Return a task status summary line for this workflow.

    Arguments:
        flow (dict):
            GraphQL JSON response for this workflow.

    Returns:
        list - Text list for the urwid.Text widget.

    """
    state_totals = flow['stateTotals']
    return [
        [
            ' ',
            (f'job_{state}', str(state_totals[state])),
            (f'job_{state}', JOB_ICON)
        ]
        for state, colour in JOB_COLOURS.items()
        if state in state_totals
        if state_totals[state]
    ]


def get_workflow_status_str(flow):
    """Return a workflow status string for the header.

    Arguments:
        flow (dict):
            GraphQL JSON response for this workflow.

    Returns:
        list - Text list for the urwid.Text widget.

    """


def _render_user(node, data):
    return f'~{ME}'


def _render_job_info(node, data):
    key_len = max(len(key) for key in data)
    ret = [
        f'{key} {" " * (key_len - len(key))} {value}\n'
        for key, value in data.items()
    ]
    ret[-1] = ret[-1][:-1]  # strip trailing newline
    return ret


def _render_job(node, data):
    return [
        f'#{data["submitNum"]:02d} ',
        get_job_icon(data['state'])
    ]


def _render_task(node, data):
    start_time = None
    mean_time = None
    try:
        # due to sorting this is the most recent job
        first_child = node.get_child_node(0)
    except IndexError:
        first_child = None

    # progress information
    if data['state'] == TASK_STATUS_RUNNING and first_child:
        start_time = first_child.get_value()['data']['startedTime']
        mean_time = data['task']['meanElapsedTime']

    # the task icon
    ret = get_task_icon(
        data['state'],
        is_held=data['isHeld'],
        is_queued=data['isQueued'],
        is_runahead=data['isRunahead'],
        start_time=start_time,
        mean_time=mean_time
    )

    # the most recent job status
    ret.append(' ')
    if first_child:
        state = first_child.get_value()['data']['state']
        ret += [(f'job_{state}', f'{JOB_ICON}'), ' ']

    # the task name
    ret.append(f'{data["name"]}')
    return ret


def _render_family(node, data):
    return [
        get_task_icon(
            data['state'],
            is_held=data['isHeld'],
            is_queued=data['isQueued'],
            is_runahead=data['isRunahead']
        ),
        ' ',
        Tokens(data['id']).pop_token()[1]
    ]


def _render_unknown(node, data):
    try:
        state_totals = get_task_status_summary(data)
        status = data['status']
        status_msg = [
            (
                'title',
                _display_workflow_id(data),
            ),
            ' - ',
            (
                f'workflow_{status}',
                status
            )
        ]
    except KeyError:
        return Tokens(data['id']).pop_token()[1]

    return [*status_msg, *state_totals]


def _display_workflow_id(data):
    return data['name']


RENDER_FUNCTIONS = {
    'user': _render_user,
    'root': _render_user,
    'job_info': _render_job_info,
    'job': _render_job,
    'task': _render_task,
    'cycle': _render_family,
    'family': _render_family,
}


def render_node(node, data, type_):
    """Render a tree node as text.

    Args:
        node (MonitorNode):
            The node to render.
        data (dict):
            Data associated with that node.
        type_ (str):
            The node type (e.g. `task`, `job`, `family`).

    """
    return RENDER_FUNCTIONS.get(type_, _render_unknown)(node, data)


def extract_context(selection):
    """Return a dictionary of all component types in the selection.

    Args:
        selection (list):
            List of element id's as extracted from the data store / graphql.

    Examples:
        >>> extract_context(['~a/b', '~a/c'])
        {'user': ['a'], 'workflow': ['b', 'c']}

        >>> extract_context(['~a/b//c/d/e']
        ... )  # doctest: +NORMALIZE_WHITESPACE
        {'user': ['a'], 'workflow': ['b'], 'cycle': ['c'],
        'task': ['d'], 'job': ['e']}

        >>> list(extract_context(['root']).keys())
        ['user']

    """
    ret = {}
    for item in selection:
        if item == 'root':
            # special handling for the Tui "root" node
            # (this represents the workflow owner which is always the same as
            # user for Tui)
            ret['user'] = ME
            continue
        tokens = Tokens(item)
        for key, value in tokens.items():
            if (
                value
                and not key.endswith('_sel')  # ignore selectors
            ):
                lst = ret.setdefault(key, [])
                if value not in lst:
                    lst.append(value)
    return ret


def get_text_dimensions(text: str) -> Tuple[int, int]:
    """Return the monospace size of a block of multiline text.

    Examples:
        >>> get_text_dimensions('foo')
        (3, 1)

        >>> get_text_dimensions('''
        ...     foo bar
        ...     baz
        ... ''')
        (11, 3)

        >>> get_text_dimensions('')
        (0, 0)

    """
    lines = text.splitlines()
    return max((0, *(len(line) for line in lines))), len(lines)



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc dump [OPTIONS] ARGS

Print state of a running workflow including tasks in the active (n=0) window.

For a more detailed view of the workflow see `cylc tui` or `cylc gui`.

For more information about a specific task, such as the current state of
its prerequisites and outputs, see 'cylc show'.

For command line monitoring:
* `cylc tui`
* `watch cylc dump -t WORKFLOW_ID` works for small simple workflows

Examples:
  # Display the state of all active tasks, sorted by cycle point:
  $ cylc dump --tasks --sort WORKFLOW_ID | grep running

  # Display the state of all active tasks in a particular cycle point:
  $ cylc dump -t WORKFLOW_ID | grep 2010082406
"""

import asyncio
import json
from typing import TYPE_CHECKING

from graphene.utils.str_converters import to_snake_case

from cylc.flow.exceptions import CylcError
from cylc.flow.id_cli import parse_id_async
from cylc.flow.option_parsers import (
    WORKFLOW_ID_ARG_DOC,
    CylcOptionParser as COP,
)
from cylc.flow.network.client_factory import get_client
from cylc.flow.terminal import cli_function

if TYPE_CHECKING:
    from optparse import Values


TASK_SUMMARY_FRAGMENT = '''
fragment tProxy on TaskProxy {
  id
  name
  cyclePoint
  state
  graphDepth
  isHeld
  isQueued
  isRunahead
  flowNums
  firstParent {
    id
  }
  jobSubmits
  jobs(sort: {keys: ["submitNum"], reverse: true}) {
    id
    state
    submitNum
    submittedTime
    startedTime
    finishedTime
    jobLogDir
    platform
    executionTimeLimit
    jobRunnerName
    jobId
  }
}
'''

FAMILY_SUMMARY_FRAGMENT = '''
fragment fProxy on FamilyProxy {
  id
  name
  cyclePoint
  state
}
'''

WORKFLOW_SUMMARY_FRAGMENT = '''
fragment wFlow on Workflow {
  name
  oldestActiveCyclePoint
  newestActiveCyclePoint
  timeZoneInfo {
    hours
    minutes
    stringBasic
    stringExtended
  }
  lastUpdated
  runMode
  states
  namespaceDefinitionOrder: nsDefOrder
  reloaded
  stateTotals
  meta {
    title
    description
    URL
    userDefined
  }
  status
  statusMsg
  families {
    name
    meta {
      title
      description
      URL
      userDefined
    }
    firstParent {
      name
    }
  }
  tasks {
    name
    meta {
      title
      description
      URL
      userDefined
    }
    meanElapsedTime
    firstParent {
      name
    }
  }
}
'''


def get_option_parser():
    parser = COP(
        __doc__,
        comms=True,
        argdoc=[WORKFLOW_ID_ARG_DOC],
    )
    parser.add_option(
        "-g", "--global", help="Global information only.",
        action="store_const", const="global", dest="disp_form")
    parser.add_option(
        "-t", "--tasks", help="Task states only.",
        action="store_const", const="tasks", dest="disp_form")
    parser.add_option(
        "-f", "--flows", help="Print flow numbers with tasks.",
        action="store_true", default=False, dest="show_flows")
    parser.add_option(
        "-r", "--raw", "--raw-format",
        help='Display raw format.',
        action="store_const", const="raw", dest="disp_form")
    parser.add_option(
        "-p", "--pretty", "--pretty-print",
        help='Display raw format with indents and newlines.',
        action="store_true", default=False, dest="pretty")
    parser.add_option(
        "-s", "--sort",
        help="Task states only; sort by cycle point instead of name.",
        action="store_true", default=False, dest="sort_by_cycle")

    return parser


@cli_function(get_option_parser)
def main(_, options: 'Values', workflow_id: str) -> None:
    asyncio.run(dump(workflow_id, options))


async def dump(workflow_id, options, write=print):
    workflow_id, *_ = await parse_id_async(
        workflow_id,
        constraint='workflows',
    )
    pclient = get_client(
        workflow_id,
        # set the default timeout a bit higher for "cylc dump" as it can
        # take a little while for Cylc to formulate the response
        timeout=options.comms_timeout or 15,
    )

    if options.sort_by_cycle:
        sort_args = {'keys': ['cyclePoint', 'name']}
    else:
        sort_args = {'keys': ['name', 'cyclePoint']}

    # retrict to the n=0 window
    graph_depth = 0

    if options.disp_form == "raw":
        query = f'''
            {TASK_SUMMARY_FRAGMENT}
            {FAMILY_SUMMARY_FRAGMENT}
            {WORKFLOW_SUMMARY_FRAGMENT}
            query ($wFlows: [ID]!, $sortBy: SortArgs) {{
              workflows (ids: $wFlows, stripNull: false) {{
                ...wFlow
                taskProxies (sort: $sortBy, graphDepth: {graph_depth}) {{
                  ...tProxy
                }}
                familyProxies (sort: $sortBy, graphDepth: {graph_depth}) {{
                  ...fProxy
                }}
              }}
            }}'''
    elif options.disp_form != "tasks":
        query = f'''
            {WORKFLOW_SUMMARY_FRAGMENT}
            query ($wFlows: [ID]!) {{
              workflows (ids: $wFlows, stripNull: false) {{
                ...wFlow
              }}
            }}'''
    else:
        query = f'''
            {TASK_SUMMARY_FRAGMENT}
            query ($wFlows: [ID]!, $sortBy: SortArgs) {{
              workflows (ids: $wFlows, stripNull: false) {{
                taskProxies (sort: $sortBy, graphDepth: {graph_depth}) {{
                  ...tProxy
                }}
              }}
            }}'''

    query_kwargs = {
        'request_string': query,
        'variables': {'wFlows': [workflow_id], 'sortBy': sort_args}
    }

    workflows = await pclient.async_request('graphql', query_kwargs)

    try:
        for summary in workflows['workflows']:
            if options.disp_form == "raw":
                if options.pretty:
                    write(json.dumps(summary, indent=4))
                else:
                    write(summary)
            else:
                if options.disp_form != "tasks":
                    node_urls = {
                        node['name']: node['meta']['URL']
                        for node in summary['tasks'] + summary['families']
                    }
                    summary['workflowUrls'] = {
                        node_name: node_urls[node_name]
                        for node_name in summary['namespaceDefinitionOrder']
                        if node_name in node_urls
                    }
                    summary['workflowUrls']['workflow_id'] = (
                        summary['meta']['URL'])
                    del summary['tasks']
                    del summary['families']
                    del summary['meta']
                    for key, value in sorted(summary.items()):
                        write(
                            f'{to_snake_case(key).replace("_", " ")}={value}')
                else:
                    for item in summary['taskProxies']:
                        if options.sort_by_cycle:
                            values = [
                                item['cyclePoint'],
                                item['name'],
                                item['state']]
                        else:
                            values = [
                                item['name'],
                                item['cyclePoint'],
                                item['state']]
                        values.append('held' if item['isHeld'] else 'not-held')
                        values.append('queued' if item['isQueued']
                                      else 'not-queued')
                        values.append('runahead' if item['isRunahead']
                                      else 'not-runahead')
                        if options.show_flows:
                            values.append(item['flowNums'])
                        write(', '.join(values))
    except Exception as exc:
        raise CylcError(
            json.dumps(workflows, indent=4) + '\n' + str(exc) + '\n')



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides base classes for cycling data objects."""

from abc import ABCMeta, abstractmethod

from cylc.flow.exceptions import CyclerTypeError


def parse_exclusion(expr: str):
    count = expr.count('!')
    if count == 0:
        return expr, None
    elif count > 1:
        raise Exception("'%s': only one set of exclusions per expression "
                        "permitted" % expr)
    else:
        remainder, exclusions = expr.split('!')
        if (
            ',' in exclusions
            and (
                not exclusions.strip().startswith('(')
                or not exclusions.strip().endswith(')')
            )
        ):
            raise Exception(
                f"'{exclusions}': a list of exclusions must be enclosed in"
                " parentheses."
            )

        exclusions = exclusions.translate(str.maketrans('', '', ' ()'))
        exclusions = exclusions.split(',')
        return remainder.strip(), exclusions


class PointBase(metaclass=ABCMeta):

    """The abstract base class for single points in a cycler sequence.

    Points should be based around a string value.

    Subclasses should provide values for TYPE and TYPE_SORT_KEY.
    They should also provide self._cmp, self.sub, self.add, and
    self.eq methods which should behave as __cmp__, __sub__,
    etc standard comparison methods. Note: "_cmp" not "cmp".

    Subclasses may also provide an overridden self.standardise
    method to reprocess their value into a standard form.

    """

    __slots__ = ('value')

    @property
    @abstractmethod
    def TYPE(self) -> str:
        """Cycling type of this point."""

    @property
    @abstractmethod
    def TYPE_SORT_KEY(self) -> int:
        """Used for comparison operations between different PointBase-derived
        class instances."""

    def __init__(self, value: str):
        if not isinstance(value, str):
            raise TypeError(type(value))
        self.value = value

    @abstractmethod
    def add(self, other):
        """Add other (interval) to self, returning a point."""
        pass

    @abstractmethod
    def _cmp(self, other) -> int:
        """Compare self to other point, returning a 'cmp'-like result."""
        pass

    def standardise(self) -> 'PointBase':
        """Format self.value into a standard representation and check it."""
        return self

    @abstractmethod
    def sub(self, other):
        """Subtract other (interval or point), returning a point or interval.

        If other is a Point, return an Interval.
        If other is an Interval, return a Point.

        ('Point' here is a PointBase-derived object, and 'Interval' an
         IntervalBase-derived object)

        """
        pass

    def __str__(self):
        # Stringify.
        return self.value

    __repr__ = __str__

    def __cmp__(self, other: 'PointBase') -> int:
        # Compare to other point.
        # TODO: refactor as __cmp__ not used in python3
        if other is None:
            return -1
        if not isinstance(other, PointBase):
            return NotImplemented
        # Note: apparently we can compare different subclasses of PointBase:
        if self.TYPE != other.TYPE:
            return cmp(self.TYPE_SORT_KEY, other.TYPE_SORT_KEY)
        if self.value == other.value:
            return 0
        return self._cmp(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.__cmp__(other) == 0
        return NotImplemented

    def __lt__(self, other: 'PointBase') -> bool:
        return self.__cmp__(other) == -1

    def __le__(self, other: 'PointBase') -> bool:
        return self.__cmp__(other) <= 0

    def __gt__(self, other: 'PointBase') -> bool:
        return self.__cmp__(other) == 1

    def __ge__(self, other: 'PointBase') -> bool:
        return self.__cmp__(other) >= 0

    def __hash__(self) -> int:
        return hash(self.value)

    def __sub__(self, other):
        # Subtract other (point or interval) from self.
        if self.TYPE != other.TYPE:
            raise CyclerTypeError(self.TYPE, self, other.TYPE, other)
        return self.sub(other)

    def __add__(self, other):
        # Add other (point or interval) from self.
        if self.TYPE != other.TYPE:
            raise CyclerTypeError(self.TYPE, self, other.TYPE, other)
        return self.add(other)


class IntervalBase(metaclass=ABCMeta):

    """An interval separating points in a cycler sequence.

    Intervals should be based around a string value.

    Subclasses should provide values for TYPE and TYPE_SORT_KEY.
    They should also provide self._cmp, self.sub, self.add,
    self.__mul__, self.__abs__ methods which should
    behave as __cmp__, __sub__, etc standard comparison methods.

    They can also just override the provided comparison methods (such
    as __cmp__) instead.

    Note: "_cmp" not "cmp", etc. They should also provide:
     * self.get_null, which is a method to extract the null interval of
    this type.
     * self.get_null_offset, which is a method to extract a null offset
    relative to a PointBase object.

    Subclasses may also provide an overridden self.standardise
    method to reprocess their value into a standard form.

    """

    __slots__ = ('value')

    @property
    @abstractmethod
    def TYPE(self) -> str:
        """Cycling type of this interval."""

    @property
    @abstractmethod
    def TYPE_SORT_KEY(self) -> int:
        """Used for comparison operations between different
        IntervalBase-derived class instances."""

    @classmethod
    @abstractmethod
    def get_null(cls):
        """Return a null interval."""
        pass

    @abstractmethod
    def __abs__(self):
        # Return an interval with absolute values for all properties.
        pass

    @abstractmethod
    def __mul__(self, factor):
        # Return an interval with all properties multiplied by factor.
        pass

    def __init__(self, value):
        if not isinstance(value, str):
            raise TypeError(type(value))
        self.value = value

    @abstractmethod
    def add(self, other):
        """Add other to self, returning a Point or Interval.

        If other is a Point, return a Point.
        If other is an Interval, return an Interval..

        ('Point' here is a PointBase-derived object, and 'Interval' an
         IntervalBase-derived object)

        """
        pass

    @abstractmethod
    def _cmp(self, other):
        """Compare self to other (interval), returning a 'cmp'-like result."""
        pass

    def standardise(self):
        """Format self.value into a standard representation."""
        return self

    @abstractmethod
    def sub(self, other):
        """Subtract other (interval) from self; return an interval."""
        pass

    def is_null(self):
        return (self == self.get_null())

    def __str__(self) -> str:
        # Stringify.
        return self.value

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self}>"

    def __add__(self, other):
        # Add other (point or interval) to self.
        if self.TYPE != other.TYPE:
            raise CyclerTypeError(self.TYPE, self, other.TYPE, other)
        return self.add(other)

    def __cmp__(self, other: 'IntervalBase') -> int:
        # Compare self to other (interval).
        # TODO: refactor as __cmp__ not used in python3
        if other is None:
            return -1
        if not isinstance(other, IntervalBase):
            return NotImplemented
        # Note: apparently we can compare different subclasses of IntervalBase:
        if self.TYPE != other.TYPE:
            return cmp(self.TYPE_SORT_KEY, other.TYPE_SORT_KEY)
        if self.value == other.value:
            return 0
        return self._cmp(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.__cmp__(other) == 0
        return NotImplemented

    def __lt__(self, other: 'IntervalBase') -> bool:
        return self.__cmp__(other) == -1

    def __le__(self, other: 'IntervalBase') -> bool:
        return self.__cmp__(other) <= 0

    def __gt__(self, other: 'IntervalBase') -> bool:
        return self.__cmp__(other) == 1

    def __ge__(self, other: 'IntervalBase') -> bool:
        return self.__cmp__(other) >= 0

    def __hash__(self) -> int:
        return hash(self.value)

    def __sub__(self, other):
        # Subtract other (interval or point) from self.
        if self.TYPE != other.TYPE:
            raise CyclerTypeError(self.TYPE, self, other.TYPE, other)
        return self.sub(other)

    def __neg__(self):
        # Return an interval with all properties multiplied by -1.
        return self * -1


class SequenceBase(metaclass=ABCMeta):

    """The abstract base class for cycler sequences.

    Subclasses should accept a sequence-specific string, a
    start context string, and a stop context string as
    constructor arguments.

    Subclasses should provide values for TYPE and TYPE_SORT_KEY.
    They should also provide get_async_expr, get_interval,
    get_offset & set_offset (deprecated), is_on_sequence,
    get_nearest_prev_point, get_next_point,
    get_next_point_on_sequence, get_first_point
    get_start_point, and get_stop_point.

    They should also provide a self.__eq__ implementation
    which should return whether a SequenceBase-derived object
    is equal to another (represents the same set of points).

    """

    __slots__ = ()

    @property
    @abstractmethod
    def TYPE(self) -> str:
        """Cycling type of this sequence."""

    @property
    @abstractmethod
    def TYPE_SORT_KEY(self) -> int:
        """Used for comparison operations between different
        SequenceBase-derived class instances."""

    @classmethod
    @abstractmethod  # Note: stacked decorator not strictly enforced in Py2.x
    def get_async_expr(cls, start_point=0):
        """Express a one-off sequence at the initial cycle point."""
        pass

    @abstractmethod
    def __init__(self, sequence_string, context_start, context_stop=None):
        """Parse sequence string according to context point strings."""
        pass

    @abstractmethod
    def get_interval(self):
        """Return the cycling interval of this sequence."""
        pass

    @abstractmethod
    def get_offset(self):
        """Deprecated: return the offset used for this sequence."""
        pass

    @abstractmethod
    def set_offset(self, i_offset):
        """Deprecated: alter state to offset the entire sequence."""
        pass

    # NOTE: not using @abstractmethod because we need to
    # patch this method for caching purposes
    def is_on_sequence(self, point):
        """Is point on-sequence, disregarding bounds?"""
        raise NotImplementedError

    @abstractmethod
    def is_valid(self, point):
        """Is point on-sequence and in-bounds?"""
        pass

    @abstractmethod
    def get_prev_point(self, point):
        """Return the previous point < point, or None if out of bounds."""
        pass

    @abstractmethod
    def get_nearest_prev_point(self, point):
        """Return the largest point < some arbitrary point."""
        pass

    @abstractmethod
    def get_next_point(self, point):
        """Return the next point > point, or None if out of bounds."""
        pass

    @abstractmethod
    def get_next_point_on_sequence(self, point):
        """Return the next point > point assuming that point is on-sequence,
        or None if out of bounds."""
        pass

    @abstractmethod
    def get_first_point(self, point):
        """Return the first point >= to point, or None if out of bounds."""
        pass

    @abstractmethod
    def get_start_point(self):
        """Return the first point of this sequence."""
        pass

    @abstractmethod
    def get_stop_point(self):
        """Return the last point of this sequence, or None if unbounded."""
        pass

    @abstractmethod
    def __eq__(self, other) -> bool:
        # Return True if other (sequence) is equal to self.
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass


class ExclusionBase(metaclass=ABCMeta):
    """A collection of points or sequences that are treated in an
    exclusionary manner"""
    __slots__ = ('exclusion_sequences', 'exclusion_points',
                 'exclusion_start_point', 'exclusion_end_point')

    def __init__(self, start_point, end_point=None):
        """creates an exclusions object that can contain integer points
        or integer sequences to be used as excluded points."""
        self.exclusion_sequences = []
        self.exclusion_points = []
        self.exclusion_start_point = start_point
        self.exclusion_end_point = end_point

    @abstractmethod
    def build_exclusions(self, excl_points):
        """Constructs the set of exclusion sequences or points"""
        pass

    def __contains__(self, point):
        """Return True if the provided point is in this exclusion.

        Args:
            point (PointBase): The cycle point to check.

        """
        if point in self.exclusion_points:
            return True
        if any(seq.is_valid(point) for seq in self.exclusion_sequences):
            return True
        return False

    def __getitem__(self, key):
        """Allows indexing of the exclusion object"""
        return self.exclusion_sequences[key]

    def __str__(self):
        returns = []
        for point in sorted(self.exclusion_points):
            returns.append(str(point))
        for sequence in self.exclusion_sequences:
            returns.append(str(sequence))
        ret = ','.join(returns)
        if ',' in ret:
            ret = '(' + ret + ')'
        return ret


def cmp(self, other):
    """Temporary replacement for the Python2 cmp function."""
    if self == other:
        return 0
    if self < other:
        return -1
    return 1
